package io.keikai.importer;

import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static io.keikai.util.Converter.asList;
import static io.keikai.util.Oid.nextOid;
import static io.keikai.util.XlsxJsonHelper.*;
import static io.keikai.util.XlsxJsonHelper.toLong;

import io.keikai.json.JSONObject;

enum ChartType {
    AREA,
    AREA_3D,
    LINE,
    LINE_3D,
    STOCK,
    RADAR,
    SCATTER,
    PIE,
    PIE_3D,
    DOUGHNUT,
    BAR,
    BAR_3D,
    COLUMN,
    COLUMN_3D,
    OF_PIE,
    SURFACE,
    SURFACE_3D,
    BUBBLE
    // TODO: excel 2016 have more chart types. like treemap..
}

class ChartSpaceJson extends ShapeJson {
    final private String _uuid;
//    final private Map _chartSpace;
//    final private Map _chartColor;
//    final private Map _chartStyle;
    // shortcut
    final private Map _chartEntry;
    final private Map _plotAreaEntry;
    private final List<ChartTypeJson> chartTypes;

    ChartSpaceJson(String uuid, Map viewAnchor, Map chartSpace,
                   Map chartColor, Map chartStyle) {
        super(viewAnchor);
        if (uuid != null) {
            _uuid = uuid;
        } else {
            _uuid = nextOid();
        }
//        _chartSpace = chartSpace;
//        _chartColor = chartColor;
//        _chartStyle = chartStyle;
        _chartEntry = $(chartSpace, "c:chartSpace", "c:chart");
        _plotAreaEntry = $x(_chartEntry, "c:plotArea");
        chartTypes = findChartTypes();
    }

    @Override
    String getName() {
        final Map map = $x(viewAnchor, "xdr:graphicFrame/xdr:nvGraphicFramePr/xdr:cNvPr");
        return map == null ? null : toStringValue(map, "name");
    }

    @Override
    String getUuid() {
        return _uuid;
    }

//    List<Map> getDateAx() {
//        return toList(_plotAreaEntry, "c:dateAx");
//    }

    List<CategoryAxisJson> getCategoryAxises() {
        final Object result = $(_plotAreaEntry, "c:catAx");
        if (result instanceof Map) {
            return asList(new CategoryAxisJson((Map) result));
        } else if (result instanceof List) {
            return ((List<Map>) result).stream().map(CategoryAxisJson::new).collect(Collectors.toList());
        } else {
            return Collections.EMPTY_LIST;
        }
    }

    List<ValueAxisJson> getValueAxises() {
        final Object result = $(_plotAreaEntry, "c:valAx");
        if (result instanceof Map) {
            return asList(new ValueAxisJson((Map) result));
        } else if (result instanceof List) {
            return ((List<Map>) result).stream().map(ValueAxisJson::new).collect(Collectors.toList());
        } else {
            return Collections.EMPTY_LIST;
        }
    }

    List<DateAxisJson> getDateAxises() {
        final Object result = $(_plotAreaEntry, "c:dateAx");
        if (result instanceof Map) {
            return asList(new DateAxisJson((Map) result));
        } else if (result instanceof List) {
            return ((List<Map>) result).stream().map(DateAxisJson::new).collect(Collectors.toList());
        } else {
            return Collections.EMPTY_LIST;
        }
    }

    boolean isSetView3D() {
        return _chartEntry.containsKey("c:view3D");
    }

    View3DJson getView3D() {
        return new View3DJson($(_chartEntry, "c:view3D"));
    }

//    boolean hasValueAxis() {
//        final Map map = $(_plotAreaEntry, "c:valAx/c:delete");
//        return map == null || toBool(map, "val") != true;
//    }
//
//    boolean hasCategoryAxis() {
//        final Map map = $(_plotAreaEntry, "c:catAx/c:delete");
//        return map == null || toBool(map, "val") != true;
//    }
//
//    String getBackgroundClr() {
//        return toStringValue($x(_chartSpace, "c:chartSpace/c:spPr/a:solidFill"), "a:schemeClr");
//    }
//
//    String getBackgroundColor() {
//        return toStringValue($x(_chartSpace, "c:chartSpace/c:spPr/a:solidFill"), "a:srgbClr");
//    }
//
//    boolean hasBorder() {
//        final result = chartSpace.jump('c:chartSpace/c:spPr/a:ln').asValue.value?.containsKey('a:noFill');
//        return result != true;
//    }
//
//    String getBorderClr() {
//        return toStringValue($x(_chartSpace, "c:chartSpace/c:spPr/a:ln/a:solidFill"), "a:schemeClr");
//    }
//
//    String getBorderColor() {
//        return toStringValue($x(_chartSpace, "c:chartSpace/c:spPr/a:ln/a:solidFill"), "a:srgbClr");
//    }

    String getDispBlanksAs() {
        return toStringValue(_chartEntry, "c:dispBlanksAs", "zero");
    }

    // KEIKAI-359
    boolean hasAutoTitleDeleted() {
        return $(_chartEntry, "c:autoTitleDeleted") != null;
    }

    // default: true
    // KEIKAI-359
    boolean isAutoTitleDeleted() {
        return toBoolean(getValue(_chartEntry, "c:autoTitleDeleted"), true);
    }


    // for bar chart only
    private boolean isColumn(Map jsonMap) {
        return "col".equals(toStringValue(jsonMap, "c:barDir"));
    }

    // default: true
    boolean isRoundedCorners() {
        return toBoolean(toValue(_chartEntry, "c:roundedCorners "), true);
    }

    //ZSS-982
    // default: true
    boolean isPlotOnlyVisibleCells() {
        return toBoolean(toValue(_chartEntry, "c:plotVisOnly"), true);
    }

    // default: true
    boolean isShowDLblsOverMax() {
        return toBoolean(toValue(_chartEntry, "c:showDLblsOverMax"), true);
    }

    // KEIKAI-359
    boolean hasTitle() {
        return $(_chartEntry, "c:title") != null;
    }

    // Tow cases:
    // 1.
    //   <c:title>
    //     <c:tx>
    //       <c:rich>
    //         <a:p>
    //           <a:r>
    //             <a:t>Foo </a:t>
    //           </a:r>
    //           <a:r>
    //             <a:t>Bar</a:t>
    //           </a:r>
    //         </a:p>
    //       </c:rich>
    //     </c:tx>
    //   </c:title>
    //
    // 2.
    //   <c:title>
    //     <c:tx>
    //       <c:rich>
    //         <a:p>
    //           <a:r>
    //             <a:t>Foo</a:t>
    //           </a:r>
    //         </a:p>
    //         <a:p>
    //           <a:r>
    //            <a:pPr>
    //              <a:defRPr/>
    //            </a:pPr>
    //            <a:endParaRPr lang="en-US"/>
    //           </a:r>
    //         </a:p>
    //       </c:rich>
    //     </c:tx>
    //   </c:title>
    // TODO rich text support
    String getTitle() {
        Object cp = $x(_chartEntry, "c:title/c:tx/c:rich/a:p");
        if (cp == null) {
            cp = $x(_chartEntry, "c:title/c:txPr/a:p");
        }
        Stream<String> stream;
        if (cp instanceof List) {
            stream = ((List<Map>) cp).stream().map(
                    (entry) -> toStringValue($(entry, "a:r"), "a:t", ""));
        } else if (cp instanceof Map) {
            stream = ((List<Map>) toList(cp, "a:r")).stream().map(
                    (entry) -> toStringValue(entry, "a:t", ""));
        } else {
            return null;
        }
        return stream.filter(Objects::nonNull).reduce("", (result, curr) -> result + curr);
    }

//    ManualLayout getLegendLayout() {
//        final Map manulaLayout = $x(_chartEntry, "c:legend/c:layout/c:manualLayout");
//        if (manulaLayout != null) {
//            return new ManualLayout(manulaLayout);
//        }
//        return null;
//    }

    LegendStyle getLegendStyle() {
        final Map legend = $(_chartEntry, "c:legend");
        if (legend != null) {
            return new LegendStyle(legend);
        }
        return null;
    }

    List<ChartTypeJson> getChartTypes() {
        return chartTypes;
    }

    private List<ChartTypeJson> findChartTypes() {
        List<ChartTypeJson> result = new ArrayList<>();
        final List<Map> pieChartList = (List<Map>) toList(_plotAreaEntry, "c:pieChart");
        if (pieChartList != null) {
            pieChartList.forEach((json) -> result.add(new ChartTypeJson(ChartType.PIE, json)));
        }
        final List<Map> pie3DChartList = (List<Map>) toList(_plotAreaEntry, "c:pie3DChart");
        if (pie3DChartList != null) {
            pie3DChartList.forEach((json) -> result.add(new ChartTypeJson(ChartType.PIE_3D, json)));
        }
        final List<Map> ofPieChartList = (List<Map>) toList(_plotAreaEntry, "c:ofPieChart");
        if (ofPieChartList != null) {
            ofPieChartList.forEach((json) -> result.add(new ChartTypeJson(ChartType.OF_PIE, json)));
        }
        final List<Map> doughnutChartList = (List<Map>) toList(_plotAreaEntry, "c:doughnutChart");
        if (doughnutChartList != null) {
            doughnutChartList.forEach((json) -> result.add(new ChartTypeJson(ChartType.DOUGHNUT, json)));
        }
        final List<Map> areaChartList = (List<Map>) toList(_plotAreaEntry, "c:areaChart");
        if (areaChartList != null){
            areaChartList.forEach((json) -> result.add(new ChartTypeJson(ChartType.AREA, json)));
        }
        final List<Map> area3DChartList = (List<Map>) toList(_plotAreaEntry, "c:area3DChart");
        if (area3DChartList != null) {
            area3DChartList.forEach((json) -> result.add(new ChartTypeJson(ChartType.AREA_3D, json)));
        }
        // bar chart should before line chart and after area chart.
        final List<Map> barChartList = (List<Map>) toList(_plotAreaEntry, "c:barChart");
        if (barChartList != null) {
            barChartList.forEach((json) -> result.add(new ChartTypeJson(
                    isColumn(json) ? ChartType.COLUMN : ChartType.BAR, json)));
        }
        final List<Map> bar3DChartList = (List<Map>) toList(_plotAreaEntry, "c:bar3DChart");
        if (bar3DChartList != null) {
            bar3DChartList.forEach((json) -> result.add(new ChartTypeJson(
                    isColumn(json) ? ChartType.COLUMN_3D : ChartType.BAR_3D, json)));
        }
        final List<Map> lineChartList =  (List<Map>) toList(_plotAreaEntry, "c:lineChart");
        if (lineChartList != null) {
            lineChartList.forEach((json) -> result.add(new ChartTypeJson(ChartType.LINE, json)));
        }
        final List<Map> line3DChartList = (List<Map>) toList(_plotAreaEntry, "c:line3DChart");
        if (line3DChartList != null) {
            line3DChartList.forEach((json) -> result.add(new ChartTypeJson(ChartType.LINE_3D, json)));
        }
        final List<Map> radarChartList = (List<Map>) toList(_plotAreaEntry, "c:radarChart");
        if (radarChartList != null) {
            radarChartList.forEach((json) -> result.add(new ChartTypeJson(ChartType.RADAR, json)));
        }
        final List<Map> scatterChartList = (List<Map>) toList(_plotAreaEntry, "c:scatterChart");
        if (scatterChartList != null) {
            scatterChartList.forEach((json) -> result.add(new ChartTypeJson(ChartType.SCATTER, json)));
        }
        final List<Map> bubbleChartList = (List<Map>) toList(_plotAreaEntry, "c:bubbleChart");
        if (bubbleChartList != null) {
            bubbleChartList.forEach((json) -> result.add(new ChartTypeJson(ChartType.BUBBLE, json)));
        }
        final List<Map> surfaceChartList = (List<Map>) toList(_plotAreaEntry, "c:surfaceChart");
        if (surfaceChartList != null) {
            surfaceChartList.forEach((json) -> result.add(new ChartTypeJson(ChartType.SURFACE, json)));
        }
        final List<Map> surface3DChartList = (List<Map>) toList(_plotAreaEntry, "c:surface3DChart");
        if (surface3DChartList != null) {
            surface3DChartList.forEach((json) -> result.add(new ChartTypeJson(ChartType.SURFACE_3D, json)));
        }
        final List<Map> stockChartList = (List<Map>) toList(_plotAreaEntry, "c:stockChart");
        if (stockChartList != null) {
            stockChartList.forEach((json) -> result.add(new ChartTypeJson(ChartType.STOCK, json)));
        }
        return result;
    }
}

class ChartTypeJson {
    final Map jsonMap;
    final ChartType chartType;

    ChartTypeJson(ChartType chartType, Map jsonMap) {
        this.chartType = chartType;
        this.jsonMap = jsonMap;
    }

    ChartType getChartType() {
        return chartType;
    }

    String getGrouping() {
        return toStringValue(jsonMap, "c:grouping");
    }

    Integer getBarOverlap() {
        return toInt(jsonMap, "c:overlap");
    }

    List<? extends SeriesJson> getSeriesList() {
        final Stream<Map> stream = ((List<Map>) toList(jsonMap, "c:ser"))
                .stream();
        switch (chartType) {
            case AREA:
            case AREA_3D:
            case LINE:
            case LINE_3D:
            case BAR:
            case BAR_3D:
            case COLUMN:
            case COLUMN_3D:
            case DOUGHNUT:
            case RADAR:
                return stream.map((data) -> new CategorySeriesJson(chartType, data)).collect(Collectors.toList());
            case PIE:
            case PIE_3D:
                return stream.map((data) -> new SingleValueSeriesJson(chartType, data)).collect(Collectors.toList());
            case BUBBLE:
                return stream.map((data) -> new XYZSeriesJson(chartType, data)).collect(Collectors.toList());
            case SCATTER:
                return stream.map((data) -> new ScatterSeriesJson(chartType, data)).collect(Collectors.toList());
            case STOCK:
            case OF_PIE:
            case SURFACE:
            case SURFACE_3D:
                //todo: StockChart, OfPieChart, SurfaceChart, Surface3DChart
            default:
                System.out.println("Not implemented yet! Chart type: [ " + chartType + "]");
                return null;
        }
    }

    List<Long> getAxIds() {
        return ((List<Map>) toList(jsonMap, "c:axId")).stream().map((v) -> toLong(v, "val"))
                .collect(Collectors.toList());
    }

    boolean isSmooth() {
        if ($(jsonMap, "c:smooth") == null) {
            return false;
        }
        return toBoolean(toValue(jsonMap, "c:smooth"), true);
    }

    Boolean hasMarker() {
        if ($(jsonMap, "c:marker") == null) {
            return null;
        }
        return toBoolean(toValue(jsonMap, "c:marker"), true);
    }
}

abstract class SeriesJson {
    private final ChartType _chartType;
    protected final Map _jsonMap;
    private final Map _titleMap;
    private final DataPointJson _dataPoint;

    SeriesJson(ChartType chartType, Map jsonMap) {
        _chartType = chartType;
        _jsonMap = jsonMap;
        _titleMap = $(_jsonMap, "c:tx");
        _dataPoint = new DataPointJson(jsonMap);
    }

    ChartType getChartType() {
        return _chartType;
    }

    boolean hasTitle() {
        return _titleMap != null;
    }

    boolean isReference() {
        return $(_titleMap, "c:strRef") != null;
    }

    String getTitleText() {
        return toStringValue(_titleMap, "c:v", "");
    }

    String getTitleFormula() {
        final Object result = getValue($(_titleMap, "c:strRef"), "c:f");
        if (result == null || result instanceof Map) { // consider: <c:f xsi:nil="true"/>
            return null;
        } else if (result instanceof String) {
            return (String) result;
        } else {
            return result.toString();
        }
    }

    int getIndex() {
        return toInt(_jsonMap, "c:idx");
    }

    int getOrder() {
        return toInt(_jsonMap, "c:order");
    }

    DataPointJson getDataPoint() {
        return _dataPoint;
    }
}

abstract class AxisJson {
    final Map _jsonMap;

    AxisJson(Map jsonMap) {
        _jsonMap = jsonMap;
    }

    Long getId() {
        return toLong(_jsonMap, "c:axId");
    }

    boolean isDelete() {
        return toBoolean(toValue(_jsonMap, "c:delete"), false);
    }

    String getPosition() {
        return toStringValue(_jsonMap, "c:axPos");
    }

    boolean hasMajorGridlines() {
        return $(_jsonMap, "c:majorGridlines") != null;
    }

    String getOrientation() {
        final Map scaling = $(_jsonMap, "c:scaling");
        return toStringValue(scaling, "c:orientation");
    }

    Double getMaximum() {
        final Map scaling = $(_jsonMap, "c:scaling");
        return toDouble(scaling, "c:max", null);
    }

    Double getMinimum() {
        final Map scaling = $(_jsonMap, "c:scaling");
        return toDouble(scaling, "c:min", null);
    }

    String getFormatCode() {
        return toStringValue($(_jsonMap, "c:numFmt"), "formatCode");
    }

    String getTickLabelPosition() {
        return toStringValue(_jsonMap, "c:tickLblPos");
    }

    Long getCrossAx() {
        return toLong(_jsonMap, "c:crossAx");
    }

    String getCrosses() {
        return toStringValue(_jsonMap, "c:crosses");
    }
}

class CategoryAxisJson extends AxisJson {
    CategoryAxisJson(Map jsonMap) {
        super(jsonMap);
    }

    String getLabelAlign() {
        return toStringValue(_jsonMap, "c:lblAlgn");
    }

    int getLabelOffset() {
        return toInt(_jsonMap, "c:lblOffset", 100);
    }
}

class ValueAxisJson extends AxisJson {
    ValueAxisJson(Map jsonMap) {
        super(jsonMap);
    }

    String getCrossBetween() {
        return toStringValue(_jsonMap, "c:crossBetween");
    }
}

class DateAxisJson extends AxisJson {
    DateAxisJson(Map jsonMap) {
        super(jsonMap);
    }

    Boolean getAuto() {
        return toBoolean(toValue(_jsonMap, "c:auto"), null);
    }

    Integer getLabelOffset() {
        return toInt(_jsonMap, "c:lblOffset");
    }

    String getBaseTimeUnit() {
        return toStringValue(_jsonMap, "c:baseTimeUnit");
    }

    Double getMajorUnit() {
        return toDouble(_jsonMap, "c:majorUnit");
    }

    String getMajorTimeUnit() {
        return toStringValue(_jsonMap, "c:majorTimeUnit");
    }

    Double getMinorUnit() {
        return toDouble(_jsonMap, "c:minorUnit");
    }

    String getMinorTimeUnit() {
        return toStringValue(_jsonMap, "c:minorTimeUnit");
    }
}

class CategorySeriesJson extends SeriesJson {
    private final DataSourceJson _categoryDataSource;
    private final DataSourceJson _valueDataSource;
    private final boolean _smooth;
	private final boolean _dataLabelsVisible; // KEIKAI-579

    CategorySeriesJson(ChartType chartType, Map jsonMap) {
        super(chartType, jsonMap);
        _categoryDataSource = new DataSourceJson($(jsonMap, "c:cat"));
        _valueDataSource = new DataSourceJson($(jsonMap, "c:val"));
        if ($(jsonMap, "c:smooth") == null) {
            // KEIKAI-305, should set to false for Excel 2007 even it is not set
            _smooth = false;
        } else {
            _smooth = toBoolean(toValue(_jsonMap, "c:smooth"), true);
        }
		// KEIKAI-579
		Object dataLabels = $(jsonMap, "c:dLbls");
		if ($((JSONObject) dataLabels, "c:delete") != null) {
			_dataLabelsVisible = !toBoolean(toValue(_jsonMap, "c:delete"), false);
		} else {
			_dataLabelsVisible = true;
		}
    }

    String getCategoryFormula() {
        return _categoryDataSource.getFormula();
    }

    String getValueFormula() {
        return _valueDataSource.getFormula();
    }

    boolean isSmooth() {
        return _smooth;
    }
	// KEIKAI-579
	boolean isDataLabelsVisible() {
		return _dataLabelsVisible;
	}
}

class XYSeriesJson extends SeriesJson {
    private final DataSourceJson _xSource;
    private final DataSourceJson _ySource;

    XYSeriesJson(ChartType chartType, Map jsonMap) {
        super(chartType, jsonMap);
        _xSource = new DataSourceJson($(this._jsonMap, "c:xVal"));
        _ySource = new DataSourceJson($(this._jsonMap, "c:yVal"));
    }

    String getXFormula() {
        return _xSource.getFormula();
    }

    String getYFormula() {
        return _ySource.getFormula();
    }
}

class ScatterSeriesJson extends XYSeriesJson {
    ScatterSeriesJson(ChartType chartType, Map jsonMap) {
        super(chartType, jsonMap);
    }

//    int getLineWidth() {
//        if (hasLine()) {
//            return UnitUtil.emuToPx(toInt($x(_jsonMap, "c:spPr/a:ln"), "w"));
//        }
//        return 0;
//    }
//
//    boolean hasLine() {
//        return $x(_jsonMap, "c:spPr/a:ln/a:solidFill") != null;
//    }
}

class XYZSeriesJson extends XYSeriesJson {
    private final DataSourceJson _zSourceJson;

    XYZSeriesJson(ChartType chartType, Map jsonMap) {
        super(chartType, jsonMap);
        final Map bubbleSizeJson = $(this._jsonMap, "c:bubbleSize");
        _zSourceJson = new DataSourceJson(bubbleSizeJson);
    }

    String getZFormatCode() {
        return _zSourceJson.getFormatCode();
    }

    String getZFormula() {
        return _zSourceJson.getFormula();
    }

}

class SingleValueSeriesJson extends CategorySeriesJson {
    SingleValueSeriesJson(ChartType chartType, Map jsonMap) {
        super(chartType, jsonMap);
    }
}

class DataPointJson {
    private final Map _jsonMap;

    DataPointJson(Map jsonMap) {
        _jsonMap = jsonMap;

    }

    private static XlsxExtractor.XlsxSchemeClrExtractor getSchemeClrExtractor(Map jsonMap, String path) {
        return new XlsxExtractor.XlsxSchemeClrExtractor($x(jsonMap, path));
    }

    private static Map<Integer, XlsxExtractor.XlsxSchemeClrExtractor> getDataPointSchemeClrExtractor(Map jsonMap, String path) {
        Map<Integer, XlsxExtractor.XlsxSchemeClrExtractor> map = new HashMap<>();
        final Object dpts = $(jsonMap, "c:dPt");
        if (dpts instanceof List) {
            ((List) dpts).forEach((each) -> {
                if (each instanceof Map) {
                    map.put(toInt((Map) each, "c:idx"), getSchemeClrExtractor((Map) each, path));
                }
            });
        } else if (dpts instanceof Map) {
            map.put(toInt((Map) dpts, "c:idx"), getSchemeClrExtractor((Map) dpts, path));
        } else {
            return Collections.EMPTY_MAP;
        }
        return map;
    }

    XlsxExtractor.XlsxSchemeClrExtractor getFillColor() {
        return getSchemeClrExtractor(_jsonMap, "c:spPr/a:solidFill");
    }

    Map<Integer, XlsxExtractor.XlsxSchemeClrExtractor> getDataPointFillColors() {
        return getDataPointSchemeClrExtractor(_jsonMap, "c:spPr/a:solidFill");
    }

    XlsxExtractor.XlsxSchemeClrExtractor getLineColor() {
        return new XlsxExtractor.XlsxSchemeClrExtractor($x(_jsonMap, "c:spPr/a:ln/a:solidFill"));
    }

    Map<Integer, XlsxExtractor.XlsxSchemeClrExtractor> getDataPointLineColors() {
        return getDataPointSchemeClrExtractor(_jsonMap, "c:spPr/a:ln/a:solidFill");
    }

    XlsxExtractor.XlsxSchemeClrExtractor getMarkerColor() {
        return getSchemeClrExtractor(_jsonMap, "c:marker/c:spPr/a:solidFill");
    }

    Map<Integer, XlsxExtractor.XlsxSchemeClrExtractor> getDataPointMarkerColors() {
        return getDataPointSchemeClrExtractor(_jsonMap, "c:marker/c:spPr/a:solidFill");
    }
}

class DataSourceJson {
    private final Map _jsonMap;
    private final boolean _isSetNumLit;
    private final boolean _isSetNumRef;
    private final boolean _isSetStrLit;
    private final boolean _isSetStrRef;

    DataSourceJson(Map jsonMap) {
        _jsonMap = jsonMap;
        _isSetNumRef = $(jsonMap, "c:numRef") != null;
        _isSetStrRef = $(jsonMap, "c:strRef") != null;
        _isSetNumLit = $(jsonMap, "c:numLit") != null;
        _isSetStrLit = $(jsonMap, "c:strLit") != null;
    }

    String getFormatCode() {
        final String path = getPath();
        if (path == null) {
            return null;
        }
        return toStringValue($(_jsonMap, path), "c:formatCode");
    }

    String getFormula() {
        return _jsonMap == null ? null : (isReference() ? getFormulaString() : getLiteralString());
    }

    private boolean isReference() {
        return _isSetNumRef || _isSetStrRef;
    }

    private String getFormulaString() {
        return toStringValue($x(_jsonMap, _isSetNumRef ? "c:numRef" : "c:strRef"), "c:f");
    }

    private String getLiteralString() {
        final StringBuilder expression = new StringBuilder("{");
        if (_isSetNumLit) {
            final List<Map> list = toList($(_jsonMap, "c:numLit"), "c:pt");
            if (list == null) {
                return null;
            }
            list.forEach((entry) -> {
                final Number v = toNum(entry, "c:v");
                expression.append((v != null ? v : 0).toString()).append(",");
            });
        } else if (_isSetStrLit) {
            final List<Map> list = toList($(_jsonMap, "c:strLit"), "c:pt");
            if (list == null) {
                return null;
            }
            list.forEach((entry) -> {
                final String v = toStringValue(entry, "c:v", "");
                expression.append("\"").append(v).append("\",");
            });
        }
        final int lastIndex = expression.length() - 1;
        if (expression.charAt(lastIndex) == ',') {
            expression.delete(lastIndex, lastIndex + 1);
        }
        expression.append("}");
        return expression.toString();
    }

    private String getPath() {
        if (_isSetNumRef) {
            return "c:numRef";
        } else if (_isSetStrRef) {
            return "c:strRef";
        } else if (_isSetNumLit) {
            return "c:numLit";
        } else if (_isSetStrLit) {
            return "c:strLit";
        }
        return null;
    }
}

class View3DJson {
    private final Map _jsonMap;

    View3DJson(Map jsonMap) {
        _jsonMap = jsonMap;
    }

    // ZSS-830
    // -90 ~ 90; default: 0
    int getRotX() {
        return Optional.ofNullable(toInt(_jsonMap, "c:rotX")).orElse(0);
    }

    // ZSS-830
    // 0 ~ 360; default: 0
    int getRotY() {
        return Optional.ofNullable(toInt(_jsonMap, "c:rotY")).orElse(0);
    }

    // ZSS-830
    // 0 ~ 240; default: 30
    int getPerspective() {
        return Optional.ofNullable(toInt(_jsonMap, "c:perspective")).orElse(30);
    }

    // ZSS-830
    // 5 ~ 500; default: 100
    int getHPercent() {
        return Optional.ofNullable(toInt(_jsonMap, "c:hPercent")).orElse(100);
    }

    // ZSS-830
    // 20 ~ 2000; default: 100
    int getDepthPercent() {
        return Optional.ofNullable(toInt(_jsonMap, "c:depthPercent")).orElse( 100);
    }

    // ZSS-830
    // default: true
    boolean isRightAngleAxes() {
        return _jsonMap == null ? true : toBool(_jsonMap, "c:rAngAx", true);
    }
}

//class ManualLayout {
//    private Map layout;
//
//    ManualLayout(Map layout) {
//        this.layout = layout;
//    }
//
//    Double getX() {
//        return toDouble(layout, "c:x");
//    }
//
//    Double getY() {
//        return toDouble(layout, "c:y");
//    }
//
//    Double getW() {
//        return toDouble(layout, "c:w");
//    }
//
//    Double getH() {
//        return toDouble(layout, "c:h");
//    }
//}

class LegendStyle {
    private Map _jsonMap;

    LegendStyle(Map legend) {
        _jsonMap = legend;
    }

    String getLegendPosition() {
        return toStringValue(_jsonMap, "c:legendPos");
    }
}


//class TitleStyle {
//    private final Map rich;
//
//    TitleStyle(Map rich) {
//        this.rich = rich;
//    }
//
//    String getFontClr() {
//        return toStringValue($x(rich, "a:p/a:pPr/a:defRPr/a:solidFill"), "a:schemeClr");
//    }
//
//    String getFontFamily() {
//        return toStringValue($x(rich, "a:p/a:r/a:rPr/a:latin"), "typeface");
//    }
//
//    String getFontColor() {
//        return toStringValue($x(rich, "a:p/a:r/a:rPr/a:solidFill"), "a:srgbClr");
//    }
//
//    int getFontSize() {
//        Map rPr = $x(rich, "a:p/a:r/a:rPr");
//        if (rPr.get("sz") != null) {
//            rPr = $x(rich, "a:p/a:pPr/a:defRPr");
//        }
//        return (int) Math.ceil(toDouble(rPr, "sz") / 100);
//    }
//
//    boolean isFontBold() {
//        Map rPr = $x(rich, "a:p/a:r/a:rPr");
//        if (rPr.get("b") != null) {
//            rPr = $x(rich, "a:p/a:pPr/a:defRPr");
//        }
//        return "1".equals(toStringValue(rPr, "b"));
//    }
//
//    boolean isFontItalic() {
//        Map rPr = $x(rich, "a:p/a:r/a:rPr");
//        if (rPr.get("i") != null) {
//            rPr = $x(rich, "a:p/a:pPr/a:defRPr");
//        }
//        return "1".equals(toStringValue(rPr, "i"));
//    }
//
//    boolean isFontUnderline() {
//        Map rPr = $x(rich, "a:p/a:r/a:rPr");
//        if (rPr.get("u") != null) {
//            rPr = $x(rich, "a:p/a:pPr/a:defRPr");
//        }
//        final String str = toStringValue(rPr, "u");
//        return str != null && !"none".equals(str);
//    }
//}
