package io.keikai.importer;

import static io.keikai.util.Oid.nextOid;
import static io.keikai.util.XlsxJsonHelper.$;
import static io.keikai.util.XlsxJsonHelper.$x;
import static io.keikai.util.XlsxJsonHelper.toInt;
import static io.keikai.util.XlsxJsonHelper.toList;
import static io.keikai.util.XlsxJsonHelper.toStringValue;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import io.keikai.model.ViewAnchor;

/**
 * ShapeJson for ShapeData only
 */
abstract class ShapeJson {
    final Map viewAnchor;

    ShapeJson(Map viewAnchor) {
        this.viewAnchor = viewAnchor;
    }

    abstract String getUuid();

    abstract String getName();

    boolean isOneCell() {
        return $(viewAnchor, "xdr:to") == null;
    }

    int getFromCol() {
        return toInt($(viewAnchor, "xdr:from"), "xdr:col");
    }

    int getFromColOff() {
        return toInt($(viewAnchor, "xdr:from"), "xdr:colOff");
    }

    int getFromRow() {
        return toInt($(viewAnchor, "xdr:from"), "xdr:row");
    }

    int getFromRowOff() {
        return toInt($(viewAnchor, "xdr:from"), "xdr:rowOff");
    }

    int getToCol() {
        return toInt($(viewAnchor, "xdr:to"), "xdr:col");
    }

    int getToColOff() {
        return toInt($(viewAnchor, "xdr:to"), "xdr:colOff");
    }

    int getToRow() {
        return toInt($(viewAnchor, "xdr:to"), "xdr:row");
    }

    int getToRowOff() {
        return toInt($(viewAnchor, "xdr:to"), "xdr:rowOff");
    }

    ViewAnchor.AnchorType getAnchorType() {
        // Ref: XSSFDrawing#createTwoCellAnchor
        final String TWO_CELL = "twoCell";
        final String ONE_CELL = "oneCell";
        final String ABSOLUTE = "absolute";
        String editAsValue = toStringValue(viewAnchor, "editAs", TWO_CELL);
        ViewAnchor.AnchorType anchorType = ViewAnchor.AnchorType.MOVE_AND_RESIZE;
        if (editAsValue.equals(ONE_CELL)) {
            anchorType = ViewAnchor.AnchorType.MOVE_DONT_RESIZE;
        } else if (editAsValue.equals(ABSOLUTE)) {
            anchorType = ViewAnchor.AnchorType.DONT_MOVE_AND_RESIZE;
        }
        return anchorType;
    }
}

/**
 * unimplemented shape json
 */

class UnimplementedShapeJson extends ShapeJson {
    //    Tuple2<String, String> hlinkClick; //Target, tooltip
    UnimplementedShapeJson(Map viewAnchor/*, this.hlinkClick*/) {
        super(viewAnchor);
    }

    @Override
    String getName() {
        return null;
    }

    @Override
    String getUuid() {
        return null;
    }
}

/**
 * Drawing Shape json
 */
class DrawingShapeJson extends ShapeJson {
    //    Tuple2<String, String> hlinkClick; //Target, tooltip
    DrawingShapeType _type;

    DrawingShapeJson(Map viewAnchor/*, this.hlinkClick*/) {
        super(viewAnchor);
    }

    @Override
    String getName() {
        return toStringValue($x(viewAnchor, "xdr:sp/xdr:nvSpPr/xdr:cNvPr"), "name");
    }

    @Override
    String getUuid() {
        return toStringValue($x(viewAnchor, "xdr:sp/xdr:nvSpPr/xdr:cNvPr"), "id");
    }

    String getText() {
        // a:p may have multiple values
        final List<Map> list = toList(viewAnchor, "xdr:sp/xdr:txBody/a:p");
        if (list == null) {
            return "";
        }
        final String v = toStringValue($(list.iterator().next(), "a:r"), "a:T");
        return v != null ? v : "";
    }

    boolean hasFill() {
        final Map map = $x(viewAnchor, "xdr:sp/xdr:spPr");
        return map != null && !map.containsKey("a:noFill");
    }

    String getHlinkTooltip() {
        final String v = toStringValue($x(viewAnchor, "xdr:sp/xdr:nvSpPr/xdr:cNvPr/a:hlinkClick"), "tooltip");
        return v == null ? "" : v;
    }

    boolean hasLine() {
        final Map map = $x(viewAnchor, "xdr:sp/xdr:spPr/a:ln");
        return map != null && !map.containsKey("a:noFill");
    }

    String getFillClr() {
        final String v = toStringValue($x(viewAnchor, "xdr:sp/xdr:spPr/a:solidFill/a:schemeClr"), "val");
        if (v != null) {
            return v;
        }
        return toStringValue($x(viewAnchor, "xdr:sp/xdr:style/a:fillRef/a:schemeClr"), "val");
    }

    String getFillAlpha() {
        final String v = toStringValue($x(viewAnchor, "xdr:sp/xdr:spPr/a:solidFill/a:schemeClr"), "a:alpha");
        if (v != null) {
            return v;
        }
        return toStringValue($x(viewAnchor, "xdr:sp/xdr:style/a:fillRef/a:schemeClr"), "a:alpha");
    }

    String getFontClr() {
        final String v = toStringValue($x(viewAnchor, "xdr:sp/xdr:txBody/a:p/a:r/a:rPr/a:solidFill/a:schemeClr"), "val");
        if (v != null) {
            return v;
        }
        return toStringValue($x(viewAnchor, "xdr:sp/xdr:style/a:fontRef/a:schemeClr"), "val");
    }

    // pt
    int getFontSize() {
        final String sz = toStringValue($x(viewAnchor, "xdr:sp/xdr:txBody/a:p/a:r/a:rPr"), "sz");
        if (sz != null) {
            return (int) Math.ceil(Double.parseDouble(sz) / 100);
        }
        return 0;
    }

    boolean isFontBold() {
        return "1".equals(toStringValue($x(viewAnchor, "xdr:sp/xdr:txBody/a:p/a:r/a:rPr"), "b"));
    }

    boolean isFontItalic() {
        return "1".equals(toStringValue($x(viewAnchor, "xdr:sp/xdr:txBody/a:p/a:r/a:rPr"), "i"));
    }

    String getFontAlign() {
        return toStringValue($x(viewAnchor, "xdr:sp/xdr:txBody/a:p/a:pPr"), "algn");
    }

    String getFontVAlign() {
        return toStringValue($x(viewAnchor, "xdr:sp/xdr:txBody/a:bodyPr"), "anchor");
    }

    String getFontFamily() {
        return toStringValue($x(viewAnchor, "xdr:sp/xdr:style/a:fontRef"), "idx");
    }

    boolean getHidden() {
        return "1".equals(toStringValue($x(viewAnchor, "xdr:sp/xdr:nvSpPr/xdr:cNvPr"), "hidden"));
    }

    DrawingShapeType getType() {
        if (_type == null) {
            _type = _getType();
        }
        return _type;
    }

    String getTypeString() {
        return toStringValue($x(viewAnchor, "xdr:sp/xdr:spPr/a:prstGeom"), "prst");
    }

    private DrawingShapeType _getType() {
        final String prst = getTypeString();
        if (prst == null) return null;
        switch (prst) {
            case "accentBorderCallout1":
                return DrawingShapeType.accentBorderCallout1;
            case "accentBorderCallout2":
                return DrawingShapeType.accentBorderCallout2;
            case "accentBorderCallout3":
                return DrawingShapeType.accentBorderCallout3;
            case "accentCallout1":
                return DrawingShapeType.accentCallout1;
            case "accentCallout2":
                return DrawingShapeType.accentCallout2;
            case "accentCallout3":
                return DrawingShapeType.accentCallout3;
            case "actionButtonBackPrevious":
                return DrawingShapeType.actionButtonBackPrevious;
            case "actionButtonBeginning":
                return DrawingShapeType.actionButtonBeginning;
            case "actionButtonBlank":
                return DrawingShapeType.actionButtonBlank;
            case "actionButtonDocument":
                return DrawingShapeType.actionButtonDocument;
            case "actionButtonEnd":
                return DrawingShapeType.actionButtonEnd;
            case "actionButtonForwardNext":
                return DrawingShapeType.actionButtonForwardNext;
            case "actionButtonHelp":
                return DrawingShapeType.actionButtonHelp;
            case "actionButtonHome":
                return DrawingShapeType.actionButtonHome;
            case "actionButtonInformation":
                return DrawingShapeType.actionButtonInformation;
            case "actionButtonMovie":
                return DrawingShapeType.actionButtonMovie;
            case "actionButtonReturn":
                return DrawingShapeType.actionButtonReturn;
            case "actionButtonSound":
                return DrawingShapeType.actionButtonSound;
            case "arc":
                return DrawingShapeType.arc;
            case "bentArrow":
                return DrawingShapeType.bentArrow;
            case "bentConnector2":
                return DrawingShapeType.bentConnector2;
            case "bentConnector3":
                return DrawingShapeType.bentConnector3;
            case "bentConnector4":
                return DrawingShapeType.bentConnector4;
            case "bentConnector5":
                return DrawingShapeType.bentConnector5;
            case "bentUpArrow":
                return DrawingShapeType.bentUpArrow;
            case "bevel":
                return DrawingShapeType.bevel;
            case "blockArc":
                return DrawingShapeType.blockArc;
            case "borderCallout1":
                return DrawingShapeType.borderCallout1;
            case "borderCallout2":
                return DrawingShapeType.borderCallout2;
            case "borderCallout3":
                return DrawingShapeType.borderCallout3;
            case "bracePair":
                return DrawingShapeType.bracePair;
            case "bracketPair":
                return DrawingShapeType.bracketPair;
            case "callout1":
                return DrawingShapeType.callout1;
            case "callout2":
                return DrawingShapeType.callout2;
            case "callout3":
                return DrawingShapeType.callout3;
            case "can":
                return DrawingShapeType.can;
            case "chartPlus":
                return DrawingShapeType.chartPlus;
            case "chartStar":
                return DrawingShapeType.chartStar;
            case "chartX":
                return DrawingShapeType.chartX;
            case "chevron":
                return DrawingShapeType.chevron;
            case "chord":
                return DrawingShapeType.chord;
            case "circularArrow":
                return DrawingShapeType.circularArrow;
            case "cloud":
                return DrawingShapeType.cloud;
            case "cloudCallout":
                return DrawingShapeType.cloudCallout;
            case "corner":
                return DrawingShapeType.corner;
            case "cornerTabs":
                return DrawingShapeType.cornerTabs;
            case "cube":
                return DrawingShapeType.cube;
            case "curvedConnector2":
                return DrawingShapeType.curvedConnector2;
            case "curvedConnector3":
                return DrawingShapeType.curvedConnector3;
            case "curvedConnector4":
                return DrawingShapeType.curvedConnector4;
            case "curvedConnector5":
                return DrawingShapeType.curvedConnector5;
            case "curvedDownArrow":
                return DrawingShapeType.curvedDownArrow;
            case "curvedLeftArrow":
                return DrawingShapeType.curvedLeftArrow;
            case "curvedRightArrow":
                return DrawingShapeType.curvedRightArrow;
            case "curvedUpArrow":
                return DrawingShapeType.curvedUpArrow;
            case "decagon":
                return DrawingShapeType.decagon;
            case "diagStripe":
                return DrawingShapeType.diagStripe;
            case "diamond":
                return DrawingShapeType.diamond;
            case "dodecagon":
                return DrawingShapeType.dodecagon;
            case "donut":
                return DrawingShapeType.donut;
            case "doubleWave":
                return DrawingShapeType.doubleWave;
            case "downArrow":
                return DrawingShapeType.downArrow;
            case "downArrowCallout":
                return DrawingShapeType.downArrowCallout;
            case "ellipse":
                return DrawingShapeType.ellipse;
            case "ellipseRibbon":
                return DrawingShapeType.ellipseRibbon;
            case "ellipseRibbon2":
                return DrawingShapeType.ellipseRibbon2;
            case "flowChartAlternateProcess":
                return DrawingShapeType.flowChartAlternateProcess;
            case "flowChartCollate":
                return DrawingShapeType.flowChartCollate;
            case "flowChartConnector":
                return DrawingShapeType.flowChartConnector;
            case "flowChartDecision":
                return DrawingShapeType.flowChartDecision;
            case "flowChartDelay":
                return DrawingShapeType.flowChartDelay;
            case "flowChartDisplay":
                return DrawingShapeType.flowChartDisplay;
            case "flowChartDocument":
                return DrawingShapeType.flowChartDocument;
            case "flowChartExtract":
                return DrawingShapeType.flowChartExtract;
            case "flowChartInputOutput":
                return DrawingShapeType.flowChartInputOutput;
            case "flowChartInternalStorage":
                return DrawingShapeType.flowChartInternalStorage;
            case "flowChartMagneticDisk":
                return DrawingShapeType.flowChartMagneticDisk;
            case "flowChartMagneticDrum":
                return DrawingShapeType.flowChartMagneticDrum;
            case "flowChartMagneticTape":
                return DrawingShapeType.flowChartMagneticTape;
            case "flowChartManualInput":
                return DrawingShapeType.flowChartManualInput;
            case "flowChartManualOperation":
                return DrawingShapeType.flowChartManualOperation;
            case "flowChartMerge":
                return DrawingShapeType.flowChartMerge;
            case "flowChartMultidocument":
                return DrawingShapeType.flowChartMultidocument;
            case "flowChartOfflineStorage":
                return DrawingShapeType.flowChartOfflineStorage;
            case "flowChartOffpageConnector":
                return DrawingShapeType.flowChartOffpageConnector;
            case "flowChartOnlineStorage":
                return DrawingShapeType.flowChartOnlineStorage;
            case "flowChartOr":
                return DrawingShapeType.flowChartOr;
            case "flowChartPredefinedProcess":
                return DrawingShapeType.flowChartPredefinedProcess;
            case "flowChartPreparation":
                return DrawingShapeType.flowChartPreparation;
            case "flowChartProcess":
                return DrawingShapeType.flowChartProcess;
            case "flowChartPunchedCard":
                return DrawingShapeType.flowChartPunchedCard;
            case "flowChartPunchedTape":
                return DrawingShapeType.flowChartPunchedTape;
            case "flowChartSort":
                return DrawingShapeType.flowChartSort;
            case "flowChartSummingJunction":
                return DrawingShapeType.flowChartSummingJunction;
            case "flowChartTerminator":
                return DrawingShapeType.flowChartTerminator;
            case "folderCorner":
                return DrawingShapeType.folderCorner;
            case "frame":
                return DrawingShapeType.frame;
            case "funnel":
                return DrawingShapeType.funnel;
            case "gear6":
                return DrawingShapeType.gear6;
            case "gear9":
                return DrawingShapeType.gear9;
            case "halfFrame":
                return DrawingShapeType.halfFrame;
            case "heart":
                return DrawingShapeType.heart;
            case "heptagon":
                return DrawingShapeType.heptagon;
            case "hexagon":
                return DrawingShapeType.hexagon;
            case "homePlate":
                return DrawingShapeType.homePlate;
            case "horizontalScroll":
                return DrawingShapeType.horizontalScroll;
            case "irregularSeal1":
                return DrawingShapeType.irregularSeal1;
            case "irregularSeal2":
                return DrawingShapeType.irregularSeal2;
            case "leftArrow":
                return DrawingShapeType.leftArrow;
            case "leftArrowCallout":
                return DrawingShapeType.leftArrowCallout;
            case "leftBrace":
                return DrawingShapeType.leftBrace;
            case "leftBracket":
                return DrawingShapeType.leftBracket;
            case "leftCircularArrow":
                return DrawingShapeType.leftCircularArrow;
            case "leftRightArrow":
                return DrawingShapeType.leftRightArrow;
            case "leftRightArrowCallout":
                return DrawingShapeType.leftRightArrowCallout;
            case "leftRightCircularArrow":
                return DrawingShapeType.leftRightCircularArrow;
            case "leftRightRibbon":
                return DrawingShapeType.leftRightRibbon;
            case "leftRightUpArrow":
                return DrawingShapeType.leftRightUpArrow;
            case "leftUpArrow":
                return DrawingShapeType.leftUpArrow;
            case "lightningBolt":
                return DrawingShapeType.lightningBolt;
            case "line":
                return DrawingShapeType.line;
            case "lineInv":
                return DrawingShapeType.lineInv;
            case "mathDivide":
                return DrawingShapeType.mathDivide;
            case "mathEqual":
                return DrawingShapeType.mathEqual;
            case "mathMinus":
                return DrawingShapeType.mathMinus;
            case "mathMultiply":
                return DrawingShapeType.mathMultiply;
            case "mathNotEqual":
                return DrawingShapeType.mathNotEqual;
            case "mathPlus":
                return DrawingShapeType.mathPlus;
            case "moon":
                return DrawingShapeType.moon;
            case "nonIsoscelesTrapezoid":
                return DrawingShapeType.nonIsoscelesTrapezoid;
            case "noSmoking":
                return DrawingShapeType.noSmoking;
            case "notchedRightArrow":
                return DrawingShapeType.notchedRightArrow;
            case "octagon":
                return DrawingShapeType.octagon;
            case "parallelogram":
                return DrawingShapeType.parallelogram;
            case "pentagon":
                return DrawingShapeType.pentagon;
            case "pie":
                return DrawingShapeType.pie;
            case "pieWedge":
                return DrawingShapeType.pieWedge;
            case "plaque":
                return DrawingShapeType.plaque;
            case "plaqueTabs":
                return DrawingShapeType.plaqueTabs;
            case "plus":
                return DrawingShapeType.plus;
            case "quadArrow":
                return DrawingShapeType.quadArrow;
            case "quadArrowCallout":
                return DrawingShapeType.quadArrowCallout;
            case "rect":
                return DrawingShapeType.rect;
            case "ribbon":
                return DrawingShapeType.ribbon;
            case "ribbon2":
                return DrawingShapeType.ribbon2;
            case "rightArrow":
                return DrawingShapeType.rightArrow;
            case "rightArrowCallout":
                return DrawingShapeType.rightArrowCallout;
            case "rightBrace":
                return DrawingShapeType.rightBrace;
            case "rightBracket":
                return DrawingShapeType.rightBracket;
            case "round1Rect":
                return DrawingShapeType.round1Rect;
            case "round2DiagRect":
                return DrawingShapeType.round2DiagRect;
            case "round2SameRect":
                return DrawingShapeType.round2SameRect;
            case "roundRect":
                return DrawingShapeType.roundRect;
            case "rtTriangle":
                return DrawingShapeType.rtTriangle;
            case "smileyFace":
                return DrawingShapeType.smileyFace;
            case "snip1Rect":
                return DrawingShapeType.snip1Rect;
            case "snip2DiagRect":
                return DrawingShapeType.snip2DiagRect;
            case "snip2SameRect":
                return DrawingShapeType.snip2SameRect;
            case "snipRoundRect":
                return DrawingShapeType.snipRoundRect;
            case "squareTabs":
                return DrawingShapeType.squareTabs;
            case "star10":
                return DrawingShapeType.star10;
            case "star12":
                return DrawingShapeType.star12;
            case "star16":
                return DrawingShapeType.star16;
            case "star24":
                return DrawingShapeType.star24;
            case "star32":
                return DrawingShapeType.star32;
            case "star4":
                return DrawingShapeType.star4;
            case "star5":
                return DrawingShapeType.star5;
            case "star6":
                return DrawingShapeType.star6;
            case "star7":
                return DrawingShapeType.star7;
            case "star8":
                return DrawingShapeType.star8;
            case "straightConnector1":
                return DrawingShapeType.straightConnector1;
            case "stripedRightArrow":
                return DrawingShapeType.stripedRightArrow;
            case "sun":
                return DrawingShapeType.sun;
            case "swooshArrow":
                return DrawingShapeType.swooshArrow;
            case "teardrop":
                return DrawingShapeType.teardrop;
            case "trapezoid":
                return DrawingShapeType.trapezoid;
            case "triangle":
                return DrawingShapeType.triangle;
            case "upArrow":
                return DrawingShapeType.upArrow;
            case "upArrowCallout":
                return DrawingShapeType.upArrowCallout;
            case "upDownArrow":
                return DrawingShapeType.upDownArrow;
            case "upDownArrowCallout":
                return DrawingShapeType.upDownArrowCallout;
            case "uturnArrow":
                return DrawingShapeType.uturnArrow;
            case "verticalScroll":
                return DrawingShapeType.verticalScroll;
            case "wave":
                return DrawingShapeType.wave;
            case "wedgeEllipseCallout":
                return DrawingShapeType.wedgeEllipseCallout;
            case "wedgeRectCallout":
                return DrawingShapeType.wedgeRectCallout;
            case "wedgeRoundRectCallout":
                return DrawingShapeType.wedgeRoundRectCallout;
            default:
                return null;
            //throw new Exception("Unknown type: $prst");
        }
    }
}

// http://officeopenxml.com/drwSp-prstGeom.php
enum DrawingShapeType {
    accentBorderCallout1,
    accentBorderCallout2,
    accentBorderCallout3,
    accentCallout1,
    accentCallout2,
    accentCallout3,
    actionButtonBackPrevious,
    actionButtonBeginning,
    actionButtonBlank,
    actionButtonDocument,
    actionButtonEnd,
    actionButtonForwardNext,
    actionButtonHelp,
    actionButtonHome,
    actionButtonInformation,
    actionButtonMovie,
    actionButtonReturn,
    actionButtonSound,
    arc,
    bentArrow,
    bentConnector2,
    bentConnector3,
    bentConnector4,
    bentConnector5,
    bentUpArrow,
    bevel,
    blockArc,
    borderCallout1,
    borderCallout2,
    borderCallout3,
    bracePair,
    bracketPair,
    callout1,
    callout2,
    callout3,
    can,
    chartPlus,
    chartStar,
    chartX,
    chevron,
    chord,
    circularArrow,
    cloud,
    cloudCallout,
    corner,
    cornerTabs,
    cube,
    curvedConnector2,
    curvedConnector3,
    curvedConnector4,
    curvedConnector5,
    curvedDownArrow,
    curvedLeftArrow,
    curvedRightArrow,
    curvedUpArrow,
    decagon,
    diagStripe,
    diamond,
    dodecagon,
    donut,
    doubleWave,
    downArrow,
    downArrowCallout,
    ellipse,
    ellipseRibbon,
    ellipseRibbon2,
    flowChartAlternateProcess,
    flowChartCollate,
    flowChartConnector,
    flowChartDecision,
    flowChartDelay,
    flowChartDisplay,
    flowChartDocument,
    flowChartExtract,
    flowChartInputOutput,
    flowChartInternalStorage,
    flowChartMagneticDisk,
    flowChartMagneticDrum,
    flowChartMagneticTape,
    flowChartManualInput,
    flowChartManualOperation,
    flowChartMerge,
    flowChartMultidocument,
    flowChartOfflineStorage,
    flowChartOffpageConnector,
    flowChartOnlineStorage,
    flowChartOr,
    flowChartPredefinedProcess,
    flowChartPreparation,
    flowChartProcess,
    flowChartPunchedCard,
    flowChartPunchedTape,
    flowChartSort,
    flowChartSummingJunction,
    flowChartTerminator,
    folderCorner,
    frame,
    funnel,
    gear6,
    gear9,
    halfFrame,
    heart,
    heptagon,
    hexagon,
    homePlate,
    horizontalScroll,
    irregularSeal1,
    irregularSeal2,
    leftArrow,
    leftArrowCallout,
    leftBrace,
    leftBracket,
    leftCircularArrow,
    leftRightArrow,
    leftRightArrowCallout,
    leftRightCircularArrow,
    leftRightRibbon,
    leftRightUpArrow,
    leftUpArrow,
    lightningBolt,
    line,
    lineInv,
    mathDivide,
    mathEqual,
    mathMinus,
    mathMultiply,
    mathNotEqual,
    mathPlus,
    moon,
    nonIsoscelesTrapezoid,
    noSmoking,
    notchedRightArrow,
    octagon,
    parallelogram,
    pentagon,
    pie,
    pieWedge,
    plaque,
    plaqueTabs,
    plus,
    quadArrow,
    quadArrowCallout,
    rect,
    ribbon,
    ribbon2,
    rightArrow,
    rightArrowCallout,
    rightBrace,
    rightBracket,
    round1Rect,
    round2DiagRect,
    round2SameRect,
    roundRect,
    rtTriangle,
    smileyFace,
    snip1Rect,
    snip2DiagRect,
    snip2SameRect,
    snipRoundRect,
    squareTabs,
    star10,
    star12,
    star16,
    star24,
    star32,
    star4,
    star5,
    star6,
    star7,
    star8,
    straightConnector1,
    stripedRightArrow,
    sun,
    swooshArrow,
    teardrop,
    trapezoid,
    triangle,
    upArrow,
    upArrowCallout,
    upDownArrow,
    upDownArrowCallout,
    uturnArrow,
    verticalScroll,
    wave,
    wedgeEllipseCallout,
    wedgeRectCallout,
    wedgeRoundRectCallout
}

class PictureJson extends ShapeJson {
    final String _uuid;
    final Map picData;
    byte[] _fileContent; // for embedded mode
    String filePath; // for external mode
    String fileType;
    final String _hlinkTarget;
    final String _hlinkTooltip;
    private boolean _external; //KEIKAI-636

    PictureJson(String uuid, Map viewAnchor, Map picData, String hlinkTarget, String hlinkTooltip) {
        super(viewAnchor);
        if (uuid != null) {
            _uuid = uuid;
        } else {
            _uuid = nextOid();
        }
        this.picData = picData;
        _hlinkTarget = hlinkTarget;
        _hlinkTooltip = hlinkTooltip;
    }

	PictureJson(String uuid, Map viewAnchor, Map picData, String filePath, String hlinkTarget,
					String hlinkTooltip, boolean isExternal) {
		this(uuid, viewAnchor, picData, hlinkTarget, hlinkTooltip);
		this.filePath = filePath;
		this._external = isExternal;
	}

    PictureJson(String uuid, Map viewAnchor, Map picData, byte[] fileContent, String fileType,
                       String hlinkTarget, String hlinkTooltip) {
        this(uuid, viewAnchor, picData, hlinkTarget, hlinkTooltip);
        this._fileContent = fileContent;
        this.fileType = fileType;
    }

    @Override
    String getName() {
        return toStringValue($x(viewAnchor, "xdr:pic/xdr:nvPicPr/xdr:cNvPr"), "name");
    }

    @Override
    String getUuid() {
        return _uuid;
    }

    byte[] getPictureData() {
        return _fileContent;
    }

	//KEIKAI-636
	boolean isExternal() {
		return _external;
	}
}

class GroupPictureJson extends ShapeJson {
    final String _uuid;
    final List<PictureJson> pictures;
    private final String _hlinkTarget;
    private final String _hlinkTooltip;

    GroupPictureJson(Map viewAnchor, String uuid, String hlinkTarget, String hlinkTooltip) {
        super(viewAnchor);
        if (uuid != null) {
            _uuid = uuid;
        } else {
            _uuid = nextOid();
        }
        _hlinkTarget = hlinkTarget;
        _hlinkTooltip = hlinkTooltip;
        pictures = new ArrayList<>();
    }

    void addPicture(PictureJson pic) {
        pictures.add(pic);
    }

    @Override
    String getName() {
        return toStringValue($x(viewAnchor, "xdr:grpSp/xdr:nvGrpSpPr/xdr:cNvPr"), "name");
    }

    @Override
    String getUuid() {
        return _uuid;
    }
}
