/* XlsxParser.java

	Purpose:
		
	Description:
		
	History:
		2:31 PM 07/02/2017, Created by jumperchen

Copyright (C) 2017 Potix Corporation. All Rights Reserved.
*/
package io.keikai.importer;

import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;
import java.util.Base64;
import java.util.HashMap;
import java.util.Map;
import java.util.function.Supplier;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import javax.xml.namespace.QName;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;

import io.keikai.importer.xml.Cell;
import io.keikai.importer.xml.JSONNode;
import io.keikai.importer.xml.Row;
import io.keikai.importer.xml.SheetData;
import io.keikai.json.JSONArray;
import io.keikai.json.JSONAware;
import io.keikai.json.JSONObject;
import org.apache.commons.io.IOUtils;
import org.jsoup.nodes.Document;
import org.jsoup.parser.ParseSettings;
import org.jsoup.parser.Parser;

/**
 * A StAX parser to convert Xlsx to JSON format.
 * <p>The syntax as follows</p>
 * <ol>
 *     <li> All are attributes:
 * <p>For example,</p>
 * <pre>{@code <foo a="a"/>}
 * </pre>
 * Result:
 * <pre><code>
 * {
 *     "foo": {
 *         "a": "a"
 *     }
 * }
 * </code></pre></li>
 * <li> Attributes and child element:
 * <p>For example,</p>
 * <pre>{@code <foo a="a">
 *     <bar>baz</bar>
 * </foo>}
 * </pre>
 * Result:
 * <pre><code>
 * {
 *     "foo": {
 * 	       "": ["a"], // store for the attr key info
 * 	       "a": "a",  // attribute
 * 	       "bar": {
 *              "_": "baz" // text only
 * 	       }
 *     }
 * }
 *</code></pre>
 * </li>
 * <li> Attributes and text node only:
 * <p>For example,</p>
 * <pre>{@code <foo a="a">bar</foo>}
 * </pre>
 * Result:
 * <pre><code>
 * {
 *     "foo": {
 *         "a": "a", // attribute
 *         "_": "baz" // text only
 *     }
 * }
 *</code></pre>
 * </li>
 * <li> Text node only:
 * <p>For example,</p>
 * <pre>{@code <foo>bar</foo>}
 * </pre>
 * Result:
 * <pre><code>
 * {
 *     "foo": {
 *         "_": "baz" // text only
 *     }
 * }
 *</code></pre>
 * </li>
 * <li> Child element and text node only:
 * <p>For example,</p>
 * <pre>{@code <foo><bar>baz</bar></foo>}
 * </pre>
 * Result:
 * <pre><code>
 * {
 *     "foo": {
 * 	       "bar": {
 *             "_": "baz" // text only
 * 	       }
 *     }
 * }
 *</code></pre>
 * </li>
 * <li> Multiple child elements with the same key:
 * <p>For example,</p>
 * <pre>{@code <foo><bar a="a"></bar><bar b="b"></bar><baz/></foo>}
 * </pre>
 * Result:
 * <pre><code>
 * {
 *     "foo": {
 *         "bar": [{
 *         	    "a": "a"
 *         	}, {
 *              "b": "b"
 *         	}],
 *         "baz": null // self-closed tag
 *      }
 * }
 *</code></pre>
 * </li>
 * </ol>
 * @author jumperchen
 */
public class XlsxParser {
	private final static Map<String,  Map<String, Supplier<JSONNode>>> _xmlSchema = new HashMap<>(1);
	private final static String EMPTY_STRING = "";
	private final static String UNDERSCORE_STRING = "_";
	private final static String XMLNS_STRING = "xmlns";
	static {
		_xmlSchema.put("worksheet", new HashMap<String, Supplier<JSONNode>>(3) {{
				put("sheetData", () -> new SheetData());
				put("row", () -> new Row());
				put("c", () -> new Cell());
			}});
	}

	public static JSONObject parse(String name, ZipInputStream zipFile)
			throws IOException, XMLStreamException {

		Map<String, InputStream> mapping = new HashMap(12);
		ZipEntry entry;
		while ((entry = zipFile.getNextEntry()) != null) {
			mapping.put(entry.getName(),
					convertZipInputStreamToInputStream(zipFile));
		}
		JSONObject json = new JSONObject();

		XMLInputFactory factory = XMLInputFactory.newInstance();
		for (Map.Entry<String, InputStream> me : mapping.entrySet()) {
			final String key = me.getKey();
			final InputStream is = me.getValue();
			if (key.endsWith(".rels") || key.endsWith(".xml")) {
				// ignore unsupported file type.
				if (key.endsWith("calcChain.xml") || key.startsWith("docProps/")) continue;

					XMLStreamReader reader = factory.createXMLStreamReader(is);
					json.put(key, parseXML(reader));

			} else if (key.endsWith(".vml")) {

				// issue #1547
				BufferedInputStream inputStream = null;
				try {
					inputStream = new BufferedInputStream(is);
					inputStream.mark(inputStream.available());
					XMLStreamReader reader = factory.createXMLStreamReader(inputStream);
					json.put(key, parseXML(reader));
				} catch (XMLStreamException e) {
					// for #1547, format the invalid content.
					Parser parser = Parser.htmlParser();

					// preserve-case
					parser.settings(new ParseSettings(true, true));

						inputStream.reset();
						Document document = parser
								.parseInput(
										IOUtils.toString(inputStream, "UTF-8"), "");
						document.outputSettings().syntax(Document.OutputSettings.Syntax.xml);
						StringReader sreader = new StringReader(document.body().html());
						XMLStreamReader reader = factory.createXMLStreamReader(sreader);
						json.put(key, parseXML(reader));

				}
			} else { // image or bin
//				Todo: Dart server is not supported yet.
				String s = Base64.getEncoder().encodeToString(IOUtils.toByteArray(is));
				json.put(key, s);
			}
			is.close();
		}

		return new JSONObject(name, json);
	}
	public static JSONObject parseXML(XMLStreamReader reader) throws XMLStreamException {
		JSONObject json = new JSONObject<>(1);

		while (reader.hasNext()) {
			int tag = reader.next();
			switch (tag) {
			case XMLStreamConstants.START_ELEMENT:
				final QName qname = reader.getName();
				final String tagName = qname.getPrefix().isEmpty() ? qname.getLocalPart()
						: qname.getPrefix() + ":" + qname.getLocalPart();
				Map<String, Supplier<JSONNode>> schemaNodes = _xmlSchema
						.get(tagName);
				parseElement(reader, json, schemaNodes);
			}
		}
		return json;
	}

	public static void parseElement(XMLStreamReader reader, JSONAware json, Map<String, Supplier<JSONNode>> schemaNodes) throws XMLStreamException {
		JSONObject pObj = json instanceof JSONObject ? (JSONObject) json : null;
		JSONNode pNode = json instanceof JSONNode ? (JSONNode) json : null;
		JSONObject attr = null;
		JSONObject cnt = null;

		// excel won't have this case
		if (pObj != null && !(pObj.isEmpty()))
			pObj.remove("_");

		final QName qname = reader.getName();
		String tagName = (qname.getPrefix().isEmpty() ? qname.getLocalPart()
				: qname.getPrefix() + ":" + qname.getLocalPart());
		Supplier<JSONNode> nodeSupplier = schemaNodes != null ? schemaNodes.get(tagName) : null;
		JSONNode cNode = null;
		if (nodeSupplier != null)
			cNode = nodeSupplier.get();

		final int nc = reader.getNamespaceCount();
		final int ac = reader.getAttributeCount();
		if (nc + ac > 0) {
			attr = new JSONObject(nc + ac);
			if (nc > 0) {
				for (int i = 0; i < nc; i++) {
					final String prefix = reader.getNamespacePrefix(i);
					final String url = reader.getNamespaceURI(i);
					attr.put((prefix == null ? XMLNS_STRING : XMLNS_STRING + ':' + prefix), url == null ? EMPTY_STRING : url);
				}
			}
			if (ac > 0) {
				for (int i = 0; i < ac; i++) {
					final QName name = reader.getAttributeName(i);
					final String value = reader.getAttributeValue(i);
					if (!name.getPrefix().isEmpty()) {
						attr.put(name.getPrefix() + ":" + name.getLocalPart(), value);
					} else {
						attr.put(name.getLocalPart(), value);
					}
				}
			}
			if (cNode != null)
				cNode.addAttributes(attr);
		}
		int state = -1;
		while (reader.hasNext()) {
			int tag = reader.next();
			switch (tag) {
			case XMLStreamConstants.END_ELEMENT:
				if (pObj != null) {
					if (cNode != null) { // has JSONNode
						if (pObj.containsKey(tagName)) {
							Object result = pObj.get(tagName);
							if (result instanceof JSONArray) {
								((JSONArray) result).add(cNode);
							} else {
								Object obj = (Object) pObj.remove(tagName);
								JSONArray jsonArray = new JSONArray();
								pObj.put(tagName, jsonArray);
								jsonArray.add(obj);
							}
						} else {
							pObj.put(tagName, cNode);
						}
					} else {
						if (pObj.containsKey(tagName)) {
							Object result = pObj.get(tagName);
							JSONObject temp = null;
							if (cnt == null) {
								temp = attr;
							} else {
								if (attr == null) {
									temp = cnt;
								} else {
									JSONArray jsonArray = new JSONArray();
									attr.forEach((k, v) -> jsonArray.add(k));
									attr.put(EMPTY_STRING, jsonArray);
									attr.putAll(cnt);
									temp = attr;
								}
							}
							if (result instanceof JSONArray) {
								((JSONArray) result)
										.add(temp);
							} else {
								JSONAware v = (JSONAware) pObj.remove(tagName);
								JSONArray array = new JSONArray();
								array.add(v);
								array.add(temp);
								pObj.put(tagName, array);
							}
						} else {
							if (cnt == null) {
								pObj.put(tagName, attr);
							} else {
								if (attr == null) {
									pObj.put(tagName, cnt);
								} else {
									JSONArray jsonArray = new JSONArray();
									attr.forEach((k, v) -> jsonArray.add(k));
									attr.put(EMPTY_STRING, jsonArray);
									attr.putAll(cnt);
									pObj.put(tagName, attr);
								}
							}
						}
					}
				} else { // JSONNode optimization
					if (cNode != null) {
						pNode.addElement(tagName, cNode);
					} else {
						if (cnt == null) {
							pNode.addElement(tagName, attr);
						} else {
							if (attr == null) {
								pNode.addElement(tagName, cnt);
							} else {
								JSONArray jsonArray = new JSONArray();
								attr.forEach((k, v) -> jsonArray.add(k));
								attr.put(EMPTY_STRING, jsonArray);
								attr.putAll(cnt);
								pNode.addElement(tagName, attr);
							}
						}
					}
				}
				return;
			case XMLStreamConstants.CHARACTERS:

				String text = reader.getText();

				// multi characters
				if (state == 2) {
					if (cNode != null) {
						cNode.addContent(text);
					} else {
						if (!text.trim().isEmpty() || cnt.isEmpty()) // tf101932131.xlsx and 09232014_153811.xlsx
							cnt.put(UNDERSCORE_STRING, cnt.get(UNDERSCORE_STRING) + text);
					}
				} else {
					state = 2;
					if (cNode != null) {
						cNode.addContent(text);
					} else {
						if (cnt == null)
							cnt = new JSONObject(3);
						if (!text.trim().isEmpty()) {
							cnt.put(UNDERSCORE_STRING, text);
						} else if (cnt.isEmpty()) {
							cnt.put(UNDERSCORE_STRING, text);
						}
					}
				}
				break;
			case XMLStreamConstants.START_ELEMENT:
				if (state == -1)
					state = 1;
				if (cNode != null) {
					parseElement(reader, cNode, schemaNodes);
				} else {
					if (cnt == null)
						cnt = new JSONObject(3);
					parseElement(reader, cnt, schemaNodes);
				}
				break;
			}
		}
	}
	private static InputStream convertZipInputStreamToInputStream(
			final ZipInputStream in) throws IOException {
		ByteArrayOutputStream out = new ByteArrayOutputStream();
		IOUtils.copy(in, out);
		InputStream is = new ByteArrayInputStream(out.toByteArray());
		return is;
	}
}
