package io.keikai.model.impl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeMap;

import io.keikai.model.CellRegion;
import io.keikai.model.SAutoFilter;
import io.keikai.model.SBook;
import io.keikai.model.sys.dependency.DependencyTable;
import io.keikai.model.sys.dependency.ObjectRef.ObjectType;
import io.keikai.model.sys.dependency.Ref;
import io.keikai.model.util.Validations;
import io.keikai.range.impl.FilterRowInfo;
import io.keikai.util.SpareSpan;

/**
 * The auto fitler implement
 * @author dennis
 * @since 3.5.0
 */
public class AutoFilterImpl extends AbstractAutoFilterAdv {
	private static final long serialVersionUID = 1L;
	
	private CellRegion _region;
	private final TreeMap<Integer,NFilterColumn> _columns;
	private Map<Integer, List<FilterRowInfo>> orderedRowInfosMap;
	// KEIKAI-366
	final private Map<Integer, SpareSpan> shallHideRows;
	private Map<Integer, Integer> filterTypes; //ZSS-1234. 1: Date, 2: Number, 3: String

	private int _lastVisible = -1;
	private int _prevLastVisible = -1;

	public AutoFilterImpl(CellRegion region){
		this._region = region;
		_columns = new TreeMap<Integer,NFilterColumn>();
		orderedRowInfosMap = new HashMap<Integer, List<FilterRowInfo>>();
		shallHideRows = new HashMap<>();
		filterTypes = new HashMap<Integer, Integer>(); //ZSS-1234
	}
	
	@Override
	public CellRegion getRegion() {
		return _region;
	}

	@Override
	public void setLastRow(int lastRow) {
		_region = new CellRegion(this._region.row, this._region.column, lastRow, this._region.lastColumn);
	}

	@Override
	public Collection<NFilterColumn> getFilterColumns() {
		return Collections.unmodifiableCollection(_columns.values());
	}

	@Override
	public NFilterColumn getFilterColumn(int index, boolean create) {
		NFilterColumn col = _columns.get(index);
		if(col==null && create){
			int s = _region.getLastColumn()-_region.getColumn()+1; 
			if(index>=s){
				throw new IllegalStateException("the column index "+index+" >= "+s);
			}
			_columns.put(index, col=new FilterColumnImpl(index));
		}
		return col;
	}

	@Override
	public void clearFilterColumn(int index) {
		_columns.remove(index);
	}
	
	@Override
	public void clearFilterColumns() {
		_columns.clear();
	}	

	//ZSS-555
	@Override
	public void renameSheet(SBook book, String oldName, String newName) {
		Validations.argNotNull(oldName);
		Validations.argNotNull(newName);
		if (oldName.equals(newName)) return; // nothing change, let go
		
		final String bookName = book.getBookName();
		// remove old ObjectRef
		Ref dependent = new ObjectRefImpl(bookName, oldName, "AUTO_FILTER", ObjectType.AUTO_FILTER);

		final DependencyTable dt = 
				((AbstractBookSeriesAdv) book.getBookSeries()).getDependencyTable();
		dt.clearDependents(dependent);
		
		// Add new ObjectRef into DependencyTable so we can extend/shrink/move
		dependent = new ObjectRefImpl(bookName, newName, "AUTO_FILTER", ObjectType.AUTO_FILTER);
		
		// prepare new dummy CellRef to enforce DataValidation reference dependency
		if (this._region != null) {
			Ref dummy = new RefImpl(bookName, newName, 
				_region.row, _region.column, _region.lastRow, _region.lastColumn);
			dt.add(dependent, dummy);
		}
	}
	
	//ZSS-688
	//@since 3.6.0
	/*package*/ AutoFilterImpl cloneAutoFilterImpl() {
		return cloneAutofilterImpl(null);
	}
	
	//ZSS-1183, ZSS-1191
	/*package*/ AutoFilterImpl cloneAutofilterImpl(SBook book) {
		final AutoFilterImpl tgt = 
				new AutoFilterImpl(new CellRegion(this._region.row, this._region.column, this._region.lastRow, this._region.lastColumn));

		for (SAutoFilter.NFilterColumn value : this._columns.values()) {
			final FilterColumnImpl srccol = (FilterColumnImpl) value;
			final FilterColumnImpl tgtcol = srccol.cloneFilterColumnImpl(book); 
			tgt._columns.put(tgtcol.getIndex(), tgtcol); //ZSS-1183
		}
		
		return tgt;
	}
	
	//ZSS-1229
	@Override
	public boolean isFiltered() {
		for (SAutoFilter.NFilterColumn value : this._columns.values()) {
			if (value.isFiltered()) {
				return true;
			}
		}
		return false;
	}

	@Override
	public int getPrevlastVisibleRow() {
		return _prevLastVisible;
	}

	@Override
	public int getLastVisibleRow() {
		return _lastVisible;
	}

	@Override
	public void setLastVisibleRow(int lastVisible) {
		_prevLastVisible = _lastVisible;
		_lastVisible = lastVisible;
	}

	//ZSS-1230
	//@since 3.9
	//@Internal
	public void putFilterColumn(int index, NFilterColumn filterColumn) {
		_columns.put(index, filterColumn);
	}
	
	//ZSS-1193, ZSS-1233: should cached in autofilter by column index
	//@since 3.9.0
	//@Internal
	//@See AutoFilterDefaultHandler
	public void setCachedSet(int index, SortedSet<FilterRowInfo> orderedRowInfos) {
		this.orderedRowInfosMap.put(index, orderedRowInfos == null ? null: 
				new ArrayList<FilterRowInfo>(orderedRowInfos));
	}
	
	//ZSS-1193, ZSS-1233: should cached in autofilter by column index
	//@since 3.9.0
	//@Internal
	//@See CustomFiltersCtrl
	public List<FilterRowInfo> getCachedSet(int index) {
		return this.orderedRowInfosMap.get(index);
	}


	// KEIKAI-366
	// In Excel, when click filter dropdown button, the values which would appeared in the list were:
	// 1. (Selected) The cell values in all the visible rows
	// 2. (Unselected) The cell values in all the hidden rows which match
	//     a. The row is hid by current filter column
	//     b. The row is not hide by other filter columns
	// To improve the performance, we use shallHideRows map to cache the row indexes which are hid by the filter column
	// when one filter column is applied. So that we don't need to re-scan when user click filter dropdown button
	// NOTE: This cache only update when a column filter is applied, which means it would be outdated in some situations
	// , e.g. the cache would not be cleared when Range.resetAutoFilter() called. As long as we access the cache only
	// when the filter column is filtered it should be fine.
	public void setShallHideCache(int index, SpareSpan shallHide) {
		shallHideRows.put(index, shallHide);
	}

	// KEIKAI-366
	public Map<Integer, SpareSpan> getShallHideCache() {
		return shallHideRows;
	}
	
	//ZSS-1234
	//@since 3.9.0
	//@Internal
	//@See CustomFiltersCtrl
	public void setFilterType(int index, int type) {
		this.filterTypes.put(index, type);
	}
	
	//ZSS-1234
	//@since 3.9.0
	//@Internal
	//@See CustomFiltersCtrl
	public int getFilterType(int index) {
		return this.filterTypes.get(index);
	}	
}
