package io.keikai.model.impl;

import io.keikai.model.*;
import io.keikai.model.util.Strings;
import io.keikai.model.util.Validations;

public class CellStyleBuilderImpl implements SCellStyleBuilder {
    final private BookImpl bookImpl;
    final private boolean inStyleTable;
    private final SCellStyle src;
    private AbstractFontAdv font;
    private FillBuilderImpl fillBuilder;
    private BorderBuilderImpl borderBuilder;
    private SCellStyle.Alignment alignment;
    private SCellStyle.VerticalAlignment verticalAlignment;
    private boolean wrapText;
    private String dataFormat;
    private boolean directFormat;
    private boolean locked;
    private boolean hidden;
    private int rotation;
    private int indention;

    CellStyleBuilderImpl(BookImpl bookImpl, SCellStyle src, boolean inStyleTable) {
        this.bookImpl = bookImpl;
        this.src = src;
        this.inStyleTable = inStyleTable;
        if (src != null) {
            alignment = src.getAlignment();
            verticalAlignment = src.getVerticalAlignment();
            wrapText = src.isWrapText();
            dataFormat = src.getDataFormat();
            directFormat = src.isDirectDataFormat();
            locked = src.isLocked();
            hidden = src.isHidden();
            rotation = src.getRotation();
            indention = src.getIndention();
        } else {
            alignment = DefaultStyle.CELL_ALIGNMENT;
            verticalAlignment = DefaultStyle.CELL_VERTICAL_ALIGNMENT;
            wrapText = DefaultStyle.CELL_WRAP_TEXT;
            dataFormat = DefaultStyle.CELL_DATA_FORMAT;
            locked = DefaultStyle.CELL_LOCKED;
            hidden = DefaultStyle.CELL_HIDDEN;
        }
    }

    @Override
    public SCellStyleBuilder font(SFont font) {
        Validations.argInstance(font, AbstractFontAdv.class);
        this.font = (AbstractFontAdv) font;
        return this;
    }

    @Override
    public SCellStyleBuilder backColor(String backColor) {
        getOrCreateFill().backColor(backColor);
        return this;
    }

    @Override
    public SCellStyleBuilder fillColor(String fillColor) {
        getOrCreateFill().fillColor(fillColor);
        return this;
    }

    @Override
    public SCellStyleBuilder fillPattern(SFill.FillPattern fillPattern) {
        getOrCreateFill().fillPattern(fillPattern);
        return this;
    }

    @Override
    public SCellStyleBuilder borderLeft(SBorder.BorderType borderLeft) {
        getOrCreateBorder().borderLeft(borderLeft);
        return this;
    }

    @Override
    public SCellStyleBuilder borderLeftColor(String color) {
        getOrCreateBorder().borderLeftColor(color);
        return this;
    }

    @Override
    public SCellStyleBuilder borderTop(SBorder.BorderType borderTop) {
        getOrCreateBorder().borderTop(borderTop);
        return this;
    }

    @Override
    public SCellStyleBuilder borderTopColor(String color) {
        getOrCreateBorder().borderTopColor(color);
        return this;
    }

    @Override
    public SCellStyleBuilder borderRight(SBorder.BorderType borderRight) {
        getOrCreateBorder().borderRight(borderRight);
        return this;
    }

    @Override
    public SCellStyleBuilder borderRightColor(String color) {
        getOrCreateBorder().borderRightColor(color);
        return this;
    }

    @Override
    public SCellStyleBuilder borderBottom(SBorder.BorderType borderBottom) {
        getOrCreateBorder().borderBottom(borderBottom);
        return this;
    }

    @Override
    public SCellStyleBuilder borderBottomColor(String color) {
        getOrCreateBorder().borderBottomColor(color);
        return this;
    }

    private BorderBuilderImpl getOrCreateBorder() {
        if (borderBuilder == null) {
            borderBuilder = new BorderBuilderImpl(inStyleTable ? bookImpl : null, src != null ? src.getBorder() : null);
        }
        return borderBuilder;
    }

    private FillBuilderImpl getOrCreateFill() {
        if (fillBuilder == null) {
            fillBuilder = new FillBuilderImpl(inStyleTable ? bookImpl : null, src != null ? src.getFill() : null);
        }
        return fillBuilder;
    }

    @Override
    public SCellStyleBuilder alignment(SCellStyle.Alignment alignment) {
        this.alignment = alignment;
        return this;
    }

    @Override
    public SCellStyleBuilder verticalAlignment(SCellStyle.VerticalAlignment verticalAlignment) {
        this.verticalAlignment = verticalAlignment;
        return this;
    }

    @Override
    public SCellStyleBuilder wrapText(boolean wrapText) {
        this.wrapText = wrapText;
        return this;
    }

    @Override
    public SCellStyleBuilder dataFormat(String dataFormat) {
        this.dataFormat = Strings.isBlank(dataFormat) ? DefaultStyle.CELL_DATA_FORMAT : dataFormat;
        return this;
    }

    @Override
    public SCellStyleBuilder directFormat(boolean directFormat) {
        this.directFormat = directFormat;
        return this;
    }

    @Override
    public SCellStyleBuilder locked(boolean locked) {
        this.locked = locked;
        return this;
    }

    @Override
    public SCellStyleBuilder hidden(boolean hidden) {
        this.hidden = hidden;
        return this;
    }

    @Override
    public SCellStyleBuilder rotation(int rotation) {
        this.rotation = rotation;
        return this;
    }

    @Override
    public SCellStyleBuilder indention(int indention) {
        this.indention = indention;
        return this;
    }

    @Override
    public SCellStyle build() {
        final AbstractFontAdv font = getFont();
        final SFill fill = getFill();
        final SBorder border = getBorder();
        if (inStyleTable) {
            final SCellStyle style = new ImmutableCellStyleImpl(
                    font == null ? bookImpl.getDefaultImmutableFont() : font.cloneFont(bookImpl), fill, border,
                    alignment, verticalAlignment, wrapText, dataFormat, directFormat,
                    locked, hidden, rotation, indention);
            return bookImpl.getOrAddCellStyle(style);
        } else {
            return new ImmutableCellStyleImpl(font == null ? bookImpl.getDefaultImmutableFont() : font, fill, border,
                    alignment, verticalAlignment, wrapText, dataFormat, directFormat,
                    locked, hidden, rotation, indention);
        }
    }

    private SBorder getBorder() {
        if (borderBuilder != null) {
            return borderBuilder.build();
        } else if (this.src != null) {
            return src.getBorder();
        }
        return null;
    }

    private SFill getFill() {
        if (fillBuilder != null) {
            return fillBuilder.build();
        } else if (this.src != null) {
            return src.getFill();
        }
        return null;
    }

    private AbstractFontAdv getFont() {
        if (font != null) {
            return font;
        } else if (this.src != null) {
            return (AbstractFontAdv) src.getFont();
        }
        return null;
    }
}
