package io.keikai.model.impl;

import io.keikai.model.SBook;
import io.keikai.model.SColor;
import org.apache.commons.codec.binary.Base64;
import org.zkoss.lang.Objects;

import javax.imageio.ImageIO;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.Optional;

public class ImmutableFillImpl extends AbstractFillAdv {
    // ZSS-1191
    public final static ImmutableFillImpl BLANK_FILL =
            new ImmutableFillImpl(DefaultStyle.FILL_PATTERN, DefaultStyle.FILL_COLOR, DefaultStyle.BACK_COLOR);
    private static final long serialVersionUID = 7543091524803612257L;
    final private SColor fillColor; // ZSS-857: default fillColor is black
    final private SColor backColor;
    final private FillPattern fillPattern;
    final private int hashCode;
    private String patternHtml; //clear cache

    // ZSS-1183
    ImmutableFillImpl(ImmutableFillImpl src, SBook book) {
        this.backColor = src.backColor == null ? null : ((AbstractColorAdv) src.backColor).cloneColor(book);
        this.fillColor = src.fillColor == null ? null : ((AbstractColorAdv) src.fillColor).cloneColor(book);
        this.fillPattern = src.fillPattern;
        hashCode = hashCode(this);
    }

    public ImmutableFillImpl(FillPattern pattern, String fgColor, String bgColor) {
        this.fillPattern = pattern;
        this.fillColor = new ColorImpl(fgColor);
        this.backColor = new ColorImpl(bgColor);
        hashCode = hashCode(this);
    }

    // ZSS-1140
    public ImmutableFillImpl(FillPattern pattern, SColor fgColor, SColor bgColor) {
        this.fillPattern = pattern;
        this.fillColor = fgColor;
        this.backColor = bgColor;
        hashCode = hashCode(this);
    }

    // use for extra fill
    ImmutableFillImpl() {
        fillColor = null;
        backColor = null;
        fillPattern = null;
        hashCode = hashCode(this);
    }

    private static String getFillPatternHtml(ImmutableFillImpl style) {
        byte[] rawData = getFillPatternBytes(style, 0, 0, 8, 4);
        StringBuilder sb = new StringBuilder();
        sb.append("background-image:url(data:image/png;base64,");
        String base64 = Base64.encodeBase64String(rawData);
        sb.append(base64).append(");");
        return sb.toString();
    }

    private static int hashCode(ImmutableFillImpl fillImpl) {
        int hash = fillImpl.getFillColor().hashCode();
        hash = hash * 31 + fillImpl.getBackColor().hashCode(); //ZSS-1145
        hash = hash * 31 + fillImpl.getFillPattern().hashCode(); //ZSS-1145
        return hash;
    }

    // ZSS-974
    public static byte[] getFillPatternBytes(ImmutableFillImpl style, int xOffset, int yOffset, int width, int height) {
        BufferedImage image = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
        Graphics2D g2 = image.createGraphics();
        // background color
        byte[] rgb = style.getBackColor().getRGB();
        g2.setColor(new Color(((int) rgb[0]) & 0xff, ((int) rgb[1]) & 0xff, ((int) rgb[2]) & 0xff));
        g2.fillRect(0, 0, width, height);
        // foreground color
        rgb = style.getFillColor().getRGB();
        g2.setColor(new Color(((int) rgb[0]) & 0xff, ((int) rgb[1]) & 0xff, ((int) rgb[2]) & 0xff));
        final byte[] patb = FillImpl.PATTERN_BYTES[style.getFillPattern().ordinal()];
        for (int y = 0; y < height; ++y) {
            final int y0 = (y + yOffset) % 4;
            byte b = patb[y0];
            if (b == 0) continue; // all zero case
            if (b == 0xff) {
                g2.drawLine(0, y, width - 1, y);
                continue;
            }
            int mask = 0x80 >>> (xOffset % 8);
            for (int x = 0; x < width; ++x) {
                if ((b & mask) != 0) {
                    g2.drawLine(x, y, x, y);
                }
                mask >>>= 1;
                if (mask == 0) mask = 0x80;
            }
        }
        ByteArrayOutputStream os = new ByteArrayOutputStream();
        try {
            ImageIO.write(image, "png", os);
        } catch (IOException e) {
            e.printStackTrace();
        } finally {
            try {
                os.close();
            } catch (IOException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
        }
        return os.toByteArray();
    }

    @Override
    public SColor getFillColor() {
        return Optional.ofNullable(fillColor).orElse(DefaultStyle.FILL_COLOR); //ZSS-1145
    }

    @Override
    public void setFillColor(SColor fillColor) {
        throw new UnsupportedOperationException("doesn't support set fill color");
    }

    @Override
    public SColor getBackColor() {
        return Optional.ofNullable(backColor).orElse(DefaultStyle.BACK_COLOR); //ZSS-1145
    }

    @Override
    public void setBackColor(SColor backColor) {
        throw new UnsupportedOperationException("doesn't support set back color");
    }

    @Override
    public FillPattern getFillPattern() {
        return Optional.ofNullable(fillPattern).orElse(DefaultStyle.FILL_PATTERN); //ZSS-1145
    }

    @Override
    public void setFillPattern(FillPattern fillPattern) {
        throw new UnsupportedOperationException("doesn't support set fill pattern");
    }

    @Override
    public boolean equals(Object other) {
        if (other == this) {
            return true;
        }
        if (!(other instanceof ImmutableFillImpl)) {
            return false;
        }
        final ImmutableFillImpl o = (ImmutableFillImpl) other;
        return Objects.equals(this.getFillColor(), o.getFillColor())  //ZSS-1145
                && Objects.equals(this.getBackColor(), o.getBackColor()) //ZSS-1145
                && Objects.equals(this.getFillPattern(), o.getFillPattern()); //ZSS-1145
    }

    @Override
    public int hashCode() {
        return this.hashCode;
    }

    public String getFillPatternHtml() {
        if (!Optional.ofNullable(patternHtml).isPresent()) {
            final FillPattern pattern = getFillPattern();
            if (pattern == FillPattern.NONE || pattern == FillPattern.SOLID) {
                patternHtml = "";
            } else {
                patternHtml = getFillPatternHtml(this);
            }
        }
        return patternHtml;
    }

    @Override
    String getStyleKey() {
        return fillPattern.ordinal() +
                "." + fillColor.getHtmlColor() +
                "." + backColor.getHtmlColor();
    }

    // ZSS-1145
    @Override
    public SColor getRawFillColor() {
        return fillColor;
    }

    // ZSS-1145
    @Override
    public SColor getRawBackColor() {
        return backColor;
    }

    // ZSS-1145
    @Override
    public FillPattern getRawFillPattern() {
        return fillPattern;
    }

    // ZSS-1183
    @Override
    ImmutableFillImpl cloneFill(SBook book) {
        return book == null ? this : new ImmutableFillImpl(this, book);
    }
}
