/* EvaluationSheetImpl.java

	Purpose:
		
	Description:
		
	History:
		Nov 8, 2013 Created by Pao Wang

Copyright (C) 2013 Potix Corporation. All Rights Reserved.
 */
package io.keikai.model.impl.sys.formula;

import java.io.Serializable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import io.keikai.model.CellRegion;
import io.keikai.model.ErrorValue;
import io.keikai.model.SBook;
import io.keikai.model.SCell;
import io.keikai.model.SCell.CellType;
import io.keikai.model.SRow;
import io.keikai.model.SSheet;
import io.keikai.model.impl.AbstractCellAdv;
import io.keikai.model.impl.AbstractRowAdv;
import io.keikai.model.impl.AbstractSheetAdv;
import io.keikai.model.impl.CellImpl;
import io.keikai.model.sys.formula.FormulaExpression;
import org.apache.poi.ss.formula.EvaluationCell;
import org.apache.poi.ss.formula.EvaluationSheet;
import org.apache.poi.ss.usermodel.FormulaError;
import org.apache.poi.ss.util.CellRangeAddress;

/**
 * 
 * @author Pao
 * @since 3.5.0
 */
public class EvalSheet implements EvaluationSheet, Serializable {
	private static final long serialVersionUID = -4811427512834490081L;
	
	private SSheet _sheet;

	public EvalSheet(SSheet sheet) {
		this._sheet = sheet;
	}
	
	public SSheet getNSheet() {
		return _sheet;
	}

	public EvaluationCell getCell(int rowIndex, int columnIndex) {
		if (rowIndex > _sheet.getEndRowIndex()) {
			return null;
		}
		SCell cell = _sheet.getCell(rowIndex, columnIndex);
		return cell.isNull() ? null : new EvalCell(this, cell);
	}

	@Override
	public Iterator<EvaluationCell> nonEmptyCellsIterator(int rowStart, int rowEnd, int columnStart, int columnEnd) {
		final List<EvaluationCell> result = new LinkedList<>();
		final Iterator<SRow> rowIterator = _sheet.getRowIterator(rowStart, rowEnd);
		while (rowIterator.hasNext()) {
			final SRow row = rowIterator.next();
			final Iterator<SCell> cellIterator = ((AbstractRowAdv) row).getCellIterator(
					false, columnStart, columnEnd);
			while (cellIterator.hasNext()) {
				final SCell cell = cellIterator.next();
				if (!cell.isNull() && cell.getType() != CellType.BLANK) {
					result.add(new EvalCell(this, cell));
				}
			}
		}
		return result.iterator();
	}

	@Override
	public Iterator<EvaluationCell> nonEmptyCellsIterator(int columnIndex) {
		final List<EvaluationCell> result = new LinkedList<>();
		final Iterator<SRow> rowIterator = _sheet.getRowIterator();

		while (rowIterator.hasNext()) {
			final SRow row = rowIterator.next();
			final SCell cell = ((AbstractRowAdv) row).getCell(columnIndex, false);
			if (cell != null && cell.getType() != CellType.BLANK) {
				result.add(new EvalCell(this, cell));
			}
		}
		return result.iterator();
	}

	public void clearAllCachedResultValues() {
		// TODO
	}

	public int getLastRowNum() {
		return _sheet.getEndRowIndex();
	}

	/*package*/ static class EvalCell implements EvaluationCell {

		private SCell cell;
		private Key key;
		private int _formulaMaxIterations = -1;
		private double _formulaMaxChange = -1;

		private EvalSheet _sheet;

		public EvalCell(EvalSheet esheet, SCell cell) {
			this.cell = cell;
			_sheet = esheet;
			SSheet sheet = cell.getSheet();
			SBook book = sheet.getBook();
			int sheetIndex = book.getSheetIndex(sheet);
			key = new Key(book.getBookName(), sheetIndex, cell.getRowIndex(), cell.getColumnIndex());
			_formulaMaxIterations = book.getFormulaMaxIterations();
			_formulaMaxChange = book.getFormulaMaxChange();
		}

		public Object getIdentityKey() {
			return key;
		}

		public EvaluationSheet getSheet() {
			return _sheet;
		}

		public int getRowIndex() {
			return cell.getRowIndex();
		}

		public int getColumnIndex() {
			return cell.getColumnIndex();
		}

		public org.apache.poi.ss.usermodel.CellType getCellType() {
			switch(cell.getType()) {
				case BLANK:
					return org.apache.poi.ss.usermodel.CellType.BLANK;
				case BOOLEAN:
					return org.apache.poi.ss.usermodel.CellType.BOOLEAN;
				case ERROR:
					return org.apache.poi.ss.usermodel.CellType.ERROR;
				case FORMULA:
					return org.apache.poi.ss.usermodel.CellType.FORMULA;
				case NUMBER:
					return org.apache.poi.ss.usermodel.CellType.NUMERIC;
				case STRING:
					return org.apache.poi.ss.usermodel.CellType.STRING;
			}
			return org.apache.poi.ss.usermodel.CellType._NONE;
		}

		public double getNumericCellValue() {
			return cell.getNumberValue().doubleValue();
		}

		public String getStringCellValue() {
			if(cell.getType() == CellType.FORMULA) {
				return cell.getFormulaValue();
			} else if(cell.getType() == CellType.STRING) {
				return cell.getStringValue();
			} else {
				return null;
			}
		}

		public boolean getBooleanCellValue() {
			return cell.getBooleanValue();
		}

		public int getErrorCellValue() {
			ErrorValue errorValue = cell.getErrorValue();
			switch(errorValue.getCode()) {
				case ErrorValue.INVALID_FORMULA:
					return FormulaError.NA.getCode(); //TODO zss 3.5 this value is not in zpoi
				default:
					return errorValue.getCode();
			}
		}

		@Override
		public CellRangeAddress getArrayFormulaRange() {
			CellRegion arrayFormulaRange = cell.getArrayFormulaRegion();
			if (arrayFormulaRange == null) {
				return null;
			}
			return CellRangeAddress.valueOf(arrayFormulaRange.getReferenceString());
		}

		@Override
		public boolean isPartOfArrayFormulaGroup() {
			return cell.isPartOfArrayFormulaGroup();
		}

		public org.apache.poi.ss.usermodel.CellType getCachedFormulaResultType() {
			return getCellType();
		}

		@Override
		public String toString() {
			return getRowIndex() + ":" + getColumnIndex() + " " + getStringCellValue();
		}

		@Override
		public int hashCode() {
			return key.hashCode();
		}

		@Override
		public boolean equals(Object obj) {
			if(this == obj)
				return true;
			if(obj == null)
				return false;
			if(!(obj instanceof EvaluationCell))
				return false;
			return key.equals(((EvaluationCell)obj).getIdentityKey());
		}
		
		//ZSS-759
		public FormulaExpression getFormulaExpression() {
			return cell.getType() == CellType.FORMULA ? 
					(FormulaExpression) ((CellImpl)cell).getValue(false) : null;
		}

		//KEIKAI-453
		/// Return the local cached numeric value of this cell
		@Override
		public Object getInitValue() {
			final Object val = ((AbstractCellAdv)cell).getInitValue();
			if(val instanceof Double){
				return (Double)val;
			}else if(val instanceof Number) {
				return ((Number) val).doubleValue();
			}else if(val instanceof Boolean) {
				return ((boolean) val) ? 1D : 0D;
			}else{
				return 0D;
			}
		}

		@Override
		public int getFormulaMaxIterations() {
			return _formulaMaxIterations;
		}

		@Override
		public double getFormulaMaxChange() {
			return _formulaMaxChange;
		}
	}

	private static class Key {
		public String bookName; //ZSS-910
		public int sheet;
		public int row;
		public int column;

		public Key(String bookName, int sheet, int row, int column) {
			this.bookName = bookName; //ZSS-910
			this.sheet = sheet;
			this.row = row;
			this.column = column;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = bookName == null ? 1 : bookName.hashCode(); //ZSS-910
			result = prime * result + column;
			result = prime * result + row;
			result = prime * result + sheet;
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if(this == obj)
				return true;
			if(obj == null)
				return false;
			if(getClass() != obj.getClass())
				return false;
			Key other = (Key)obj;
			if(bookName != other.bookName || (bookName != null && !bookName.equals(other.bookName))) //ZSS-910
				return false;
			if(column != other.column)
				return false;
			if(row != other.row)
				return false;
			if(sheet != other.sheet)
				return false;
			return true;
		}

	}

	//ZSS-596 Possible memory leak when formula evaluation
	//implement hashCode and equals, use identity to implement equals
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((_sheet == null) ? 0 : _sheet.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		EvalSheet other = (EvalSheet) obj;
		
		return _sheet == other._sheet;
	}

	//ZSS-962
	@Override
	public boolean isHidden(int rowIndex, int columnIndex) {
		return ((AbstractSheetAdv)_sheet).isHidden(rowIndex, columnIndex);
	}

	// KEIKAI-63
	@Override
	public boolean isRowHidden(int rowIndex) {
		return ((AbstractSheetAdv)_sheet).isRowHidden(rowIndex);
	}

	// KEIKAI-63
	@Override
	public boolean isFiltered() {
		return ((AbstractSheetAdv)_sheet).isFiltered();
	}
}
