package io.keikai.range.impl.imexp;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import io.keikai.model.SCellStyle;
import io.keikai.model.SExtraStyle;
import io.keikai.model.SFont;
import org.apache.poi.ss.usermodel.CellStyle;

public class StyleCache {
    final private CellStyleCache cellStyleCache = new CellStyleCache();
    final private ExtraStyleCache extraStyleCache = new ExtraStyleCache();
    final private FontCache fontCache = new FontCache();

    public CellStyleCache getCellStyleCache() {
        return cellStyleCache;
    }

    public ExtraStyleCache getExtraStyleCache() {
        return extraStyleCache;
    }

    FontCache getFontCache() {
        return fontCache;
    }

    public void clear() {
        cellStyleCache.clear();
        extraStyleCache.clear();
        fontCache.clear();
    }

    /**
     * The 2-key cache of the mapping between imported {@link CellStyle} and {@link SCellStyle} to avoid the cost of creating redundant {@link SCellStyle} (consume memory) and comparing the same {@link CellStyle} (equals() costs high). <br/>
     * There are 2 keys for each {@link SCellStyle}:
     * <ol>
     * <li>{@link CellStyle}'s index </li>
     * <li>{@link CellStyle} object</li>
     * </ol>
     * <p>
     * The algorithm is: <br/>
     * <ol>
     * <li>Get {@link SCellStyle} by {@link CellStyle}'s index first, if it exists, just return</li>
     * <li>If nothing found in the previous step, get by key {@link CellStyle}, if it exists, just return. So that we can avoid creating redundant {@link SCellStyle}</li>. Because 2 styles with different indexes can have identical content (ZSS-685). <br/>
     * <li>If we can't find {@link SCellStyle} in the cache, we store the created {@link SCellStyle} with 2 keys above.</li>
     * </ol>
     */
     static class CellStyleCache {
        private final Map<Integer, SCellStyle> indexMap = new HashMap<>();
        private final Map<CellStyle, SCellStyle> cellStyleMap = new HashMap<>(); //ZSS-685

        public SCellStyle get(CellStyle poiCellStyle) {
            SCellStyle cellStyle = indexMap.get(poiCellStyle.getIndex());
            if (cellStyle == null) {
                cellStyle = this.cellStyleMap.get(poiCellStyle);
            }
            return cellStyle;
        }

        public void put(CellStyle poiCellStyle, SCellStyle zssCellStyle) {
            indexMap.put((int) poiCellStyle.getIndex(), zssCellStyle);
            cellStyleMap.put(poiCellStyle, zssCellStyle);
        }

        public void clear() {
            indexMap.clear();
            cellStyleMap.clear();
        }
    }

    public static class ExtraStyleCache {
        final private List<SExtraStyle> extraStyleCache = new ArrayList<>();

        public SExtraStyle get(int index) {
            return extraStyleCache.get(index);
        }

        void clear() {
            extraStyleCache.clear();
        }

        public void add(SExtraStyle extraStyle) {
            extraStyleCache.add(extraStyle);
        }
    }

    public static class FontCache {
        /**
         * <poi Font index, {@link SFont} object>
         **/
        final private Map<Integer, SFont> fontMap = new HashMap<>();

        SFont get(int index) {
            return fontMap.get(index);
        }

        void put(Integer index, SFont font) {
            fontMap.put(index, font);
        }

        void clear() {
            fontMap.clear();
        }
    }
}