/* SpareSpan.java

	Purpose:
		
	Description:
		The implementation refers to SimpleSpareSpan in Keikai dart version.
	History:
		10:27 AM 2019/11/22, Created by jumperchen

Copyright (C) 2019 Potix Corporation. All Rights Reserved.
*/
package io.keikai.util;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.List;
import java.util.stream.Stream;

/**
 * @author jumperchen
 */
public class SpareSpan {
	private int start;
	private int end;
	private List<Span> spans = new ArrayList();

	public boolean isEmpty() {
		return spans.isEmpty();
	}

	public int size() {
		return spans.size();
	}

	public Stream<Span> stream() {
		return spans.stream();
	}

	public void clear() {
		spans.clear();
	}

	private static Comparator _ReadSpanComparator = new Comparator<Span>() {
		public int compare(Span inSpansSpan, Span keySpan) {
			int key = keySpan.getStart();
			return key < inSpansSpan.getStart() ? -1 : key >= inSpansSpan.getEnd() ? 1 : 0;
		}
	};

	private static Comparator _SplitSpanComparator = new Comparator<Span>() {
		public int compare(Span inSpansSpan, Span keySpan) {
			int key = keySpan.getStart();
			return key <= inSpansSpan.getStart() ? -1 : key >= inSpansSpan.getEnd() ? 1 : 0;
		}
	};

	private static Comparator _WriteSpanComparator = new Comparator<Span>() {
		public int compare(Span inSpansSpan, Span keySpan) {
			int key = keySpan.getStart();
			return key < inSpansSpan.getStart() ? -1 : key > inSpansSpan.getEnd() ? 1 : 0;
		}
	};

	private static Comparator _MergeSpanComparator = _WriteSpanComparator;

	// cannot use Collections.binarySearch(), because its result is not what we want.
	private static <T> int intBinarySearch(List<? extends T> l, T key, Comparator<? super T> c) {
		int low = 0;
		int hgh = l.size()-1;

		while (low <= hgh) {
			int mid = (low + hgh) >>> 1;
			T midVal = l.get(mid);
			int cmp = c.compare(midVal, key);

			if (cmp == 0) {
				return mid;
			} else if (cmp < 0) {
				hgh = mid - 1;
			} else {
				low = mid + 1;
			}
		}
		return -(low + 1);
	}
	public Span getSpan(int index) {
		int j = intBinarySearch(spans, new Span(index, 1), _ReadSpanComparator);
		return j < 0 ? null : spans.get(j);
	}

	public int getStart() {
		return start;
	}

	public int getEnd() {
		return end;
	}

	/**
	 * Split the span at index and extends the index with the size.
	 */
	public void extend(int index, int size) {
		int j = intBinarySearch(spans, new Span(index, 1), _SplitSpanComparator);
		final int len = spans.size();
		if (j < 0) {
			for (int k = -j - 1; k < len;++k) {
				final Span span0 = spans.get(k);
				span0.start += size;
			}
		} else {
			final Span span = spans.get(j);
			final Span span2 = span.split(index);
			j += 1;
			spans.add(j, span2);
			for (int k = j; k <= len; ++k) {
				final Span span1 = spans.get(k);
				span1.start += size;
			}
		}
	}

	/**
	 * Split the span at index and extends the index with the size.
	 */
	public void extendAndMerge(int index, int size) {
		int j =  intBinarySearch(spans, new Span(index, 1), _MergeSpanComparator);
		final int len = spans.size();
		if (j < 0) {
			for (int k = -j - 1; k < len; ++k) {
				final Span span0 = spans.get(k);
				span0.start += size;
			}
		} else {
			final Span span = spans.get(j);
			span.length = span.length + size;
			j += 1;
			final int len0 = spans.size();
			for (int k = j; k < len0; ++k) {
				final Span span0 = spans.get(k);
				span0.start += size;
			}
		}
	}

	/**
	 * Shrink spans that is covered from [startIdx] to [startIdx] + [size] - 1
	 */
	public void shrink(int startIdx, int size) {
		final List<Integer> rng = removeSpans(startIdx, size);
		final Integer sj = rng.get(0);

		if (sj != null) {
			final int len = spans.size();
			final int size0 = -size;
			for (int k = sj; k < len; ++k) {
				final Span span = spans.get(k);
				span.start += size0;
			}
		}
	}

	// return a tuple of <span index after the removed spans, removed spans>
	public List<Integer> removeSpans(int startIdx, int size) {
		final int endIdx = startIdx + size;
		int sj = intBinarySearch(spans, new Span(startIdx, 1), _SplitSpanComparator);
		int ej = intBinarySearch(spans, new Span(endIdx, 1), _SplitSpanComparator);

		if (ej < 0) {
			ej = -ej - 1;
		} else {
			final Span span = spans.get(ej);
			final Span span2 = span.split(endIdx);
			ej += 1;
			spans.add(ej, span2);
		}

		if (sj < 0) {
			sj = -sj - 1;
		} else {
			final Span span = spans.get(sj);
			span.split(startIdx);
			sj += 1;
		}
		if (sj < ej) {
			spans.subList(sj, ej).clear();
		}
		return Arrays.asList(sj, ej);
	}

	// return the span at the specified span index
	public Span getSpanAt(int sj) {
		return spans.get(sj);
	}

	// delete the span at the specified span index
	public Span deleteSpanAt(int sj) {
		return spans.remove(sj);
	}
	// return the start span index(inclusive) and end span index(exclusive) of the
	// affected spans.
	public void addSpan(int start, int length) {
		_addSpanMerge(start, length);
	}

	public void deleteSpan(int start, int length) {
		// TODO: should not split the spans between the start and the length, just delete them directly.
		// The implementation is the same as SimpleSpareSpan in dart version.
		List<Integer> affected = _addSpanSplit(start, length);
		for (int s = affected.get(0), it = s, e = affected.get(1); it < e; it++) {
			deleteSpanAt(s);
		}
	}


	// merge spans into one span if the new added span glue them together.
	//
	// ```
	//  |-----|          |-----|           |---|  |------|   |--| |--| ... |--|
	//     |-------|           |----|         |-----|          |-------~~~---|
	//
	//
	//  |----------|     |----------|      |-------------|   |--------~~~-----|
	// ```
	private List<Integer> _addSpanMerge(int start, int length) {
		final int end = start + length;
		int sj = intBinarySearch(spans, new Span(start, 1), _WriteSpanComparator);
		int ej = intBinarySearch(spans, new Span(end, 1), _WriteSpanComparator);

		if (sj == ej) {
			if (sj < 0) {
				sj = -sj - 1;
				spans.add(sj, new Span(start, length));
			}
			return Arrays.asList(sj, sj + 1);
		}

		int s, e;
		if (sj < 0) {
			s = start;
			sj = -sj - 1;
		} else {
			s = spans.get(sj).start;
		}
		if (ej < 0) {
			e = end;
			ej = -ej - 1;
		} else {
			e = spans.get(ej).getEnd();
			ej += 1;
		}
		if (sj < ej) {
			spans.subList(sj, ej).clear();
		}
		spans.add(sj, new Span(s, e - s));
		return Arrays.asList(sj, sj + 1);
	}

	// split spans if the new added span cut them
	//
	// ```
	//  |-----|          |-----|           |----|  |------|
	//     |-------|           |----|         |------|
	//
	//
	//  |--|--|----|     |-----|----|      |--|-|--|-|----|
	// ```
	private  List<Integer> _addSpanSplit(int start, int length) {
		final int end = start + length;
		int sj = intBinarySearch(spans, new Span(start, 1), _SplitSpanComparator);
		int ej = intBinarySearch(spans, new Span(end, 1), _SplitSpanComparator);
		int end0 = end, start0 = start;
		int asj, aej;
		if (ej < 0) {
			ej = -ej - 1;
			aej = ej;
		} else {
			final Span span1 = spans.get(ej);
			final Span span2 = span1.split(end);
			spans.add(ej + 1, span2);
			aej = ej + 1;
			end0 = span1.start;
		}

		if (sj < 0) {
			sj = -sj - 1;
			asj = sj;
		} else {
			final Span span1 = spans.get(sj);
			final Span span2 = span1.split(start);
			sj += 1;
			spans.add(sj, span2);
			asj = sj;
			sj += 1;
			ej += 1;
			aej += 1;
			start0 = span2.getEnd();
		}

		int j = ej;
		while (--j >= sj) {
			final Span span1 = spans.get(j);
			final int end1 = span1.getEnd();
			if (end1 < end0) {
				spans.add(j+1, new Span(end1, end0 - end1));
				aej += 1;
			}
			end0 = span1.start;
		}
		if (start0 < end0) {
			spans.add(j+1, new Span(start0, end0 - start0));
			aej += 1;
		}

		return Arrays.asList(asj, aej);
	}
//
//	public static void main(String[] args) {
//		SpareSpan spareSpan = new SpareSpan();
//		for (int i = 0; i < 20; i++) {
//			if (i % 3  != 0)
//				spareSpan.addSpan(i, 1);
//		}
//		System.out.println(spareSpan.size());
//		System.out.println(spareSpan.stream().map((s) -> s.toString()).collect(Collectors.joining()));
//
//		spareSpan.deleteSpan(5, 8);
//		System.out.println(spareSpan.stream().map((s) -> s.toString()).collect(Collectors.joining()));
//	}
}
