/* StyleSheet.js

	Purpose:
        generate CSS class and its rules into a <style> in <head>
	Description:

	History:
		Wed Jan 22 15:01:44 CST 2020, Created by jumperchen

Copyright (C) 2020 Potix Corporation. All Rights Reserved.
*/
(function () {
    // simulate Java StringBuffer API
    function StringBuffer() {
        this._str = '';
        this.append = function (str) {
            this._str += str;
            return this;
        };
        this.toString = function () {
            return this._str;
        }
    }
zss.StyleSheet = zk.$extends(zk.Object, {
    $init: function (sheet, uuid) {
        this.$supers('$init', arguments);
        this._sheet = sheet;
        this._uuid = uuid;
        this._initStyle();
    },
    destroy: function () {
        if (this._node) {
            jq(this._node).remove();
        }
        this._node = null;
    },
    _initOrApplyCSS: function (css) {
        if (!this._node) {
            var head = document.head || document.getElementsByTagName('head')[0],
                style = document.createElement('style');

            head.appendChild(style);

            style.type = 'text/css';
            style.id = this._uuid;
            this._node = style;
        }
        this._node.href = this._sheet._wgt.getSheetId(); //zcss.js uses it
        if (this._node.styleSheet){
            // This is required for IE8 and below.
            this._node.styleSheet.cssText = css;
        } else {
            jq(this._node).empty();
            this._node.appendChild(document.createTextNode(css));
        }
    },
    rebuild: function () {
      this._initStyle();
    },
    _initStyle: function () {
        let sb = new StringBuffer();
        this._prepareBasicStyleSheet(sb);
        this._prepareColHeaderStyleSheet(sb);
        this._prepareRowHeaderStyleSheet(sb);
        this._prepareMergeStyleSheet(sb);
        this._prepareOtherStyleSheet(sb);
        this._initOrApplyCSS(sb.toString());
    },
    _prepareBasicStyleSheet: function (sb) {
        let _spreadsheet = this._sheet._wgt;

        let colHelper = this._sheet.custColWidth;
        let rowHelper = this._sheet.custRowHeight;

        let hiderow = _spreadsheet.isRowHeadHidden();
        let hidecol = _spreadsheet.isColumnHeadHidden();
        let showgrid = _spreadsheet.isDisplayGridlines();

        // ZSS-938 set th as 0 is for hiding 1-pixel-height part of column header
        // in case, i didn't find any inpropriate look for using 1px to row header however.
        // so i decide to only adjust column header.
        // KEIKAI-195: remove redundant borders
        let th = hidecol ? 0 : _spreadsheet.getTopheadHeight();
		let lw = hiderow ? 0 : _spreadsheet.getLeftheadWidth();
        let cp = _spreadsheet.getCellPadding();
        let rh = _spreadsheet.getRowHeight();
        let cw = _spreadsheet.getColumnWidth();
        let lh = 20;// default line height;

        if (lh > rh) {
            lh = rh;
        }

        let sheetPrefix = " .s" + _spreadsheet.getSheetId();
        let name = "#" + _spreadsheet.uuid;

        let cellwidth;// default
        let cellheight;// default

        cellwidth = cw;
        cellheight = rh;

        let celltextwidth = cw - 2 * cp;

        // zcss.setRule(name+" .zsdata",["padding-top","padding-left"],[th+"px",lw+"px"],true,sid);
        sb.append(name).append(" .zsdata{");
        //ZSS-948: + 1 to avoid 1st row's top border covered by heading's bottom border
        //		sb.append("padding-top:").append(th).append("px;");
        // KEIKAI-215
        sb.append("padding-top:").append(th).append("px;");
        sb.append("padding-left:").append(lw).append("px;");
        sb.append("}");

        // zcss.setRule(name+" .zsrow","height",rh+"px",true,sid);
        sb.append(name).append(" .zsrow{");
        sb.append("height:").append(rh).append("px;");
        sb.append("}");

        // zcss.setRule(name+" .zscell",["padding","height","width","line-height"],["0px "+cp+"px 0px "+cp+"px",cellheight+"px",cellwidth+"px",lh+"px"],true,sid);
        sb.append(name).append(" .zscell{");
        sb.append("padding:").append("0px " + cp + "px 0px " + cp + "px;");
        sb.append("height:").append(cellheight).append("px;");
        sb.append("width:").append(cellwidth).append("px;");
        if (!showgrid) {
            sb.append("border-bottom:1px solid #FFFFFF;")
                .append("border-right:1px solid #FFFFFF;");
        }
        // sb.append("line-height:").append(lh).append("px;\n");
        sb.append("}");

        // zcss.setRule(name+" .zscelltxt",["width","height"],[celltextwidth+"px",cellheight+"px"],true,sid);
        sb.append(name).append(" .zscelltxt{");
        sb.append("width:").append(celltextwidth).append("px;");
        sb.append("height:").append(cellheight).append("px;");
        sb.append("}");

        // zcss.setRule(name+" .zstop",["left","height","line-height"],[lw+"px",(th-2)+"px",lh+"px"],true,sid);

        let toph = th;
        let topheadh = toph;
        let cornertoph = th;
        // let topblocktop = toph;
        let fzr = _spreadsheet.getRowFreeze();
        let fzc = _spreadsheet.getColumnFreeze();

        if (fzr > -1) {
            toph = toph + rowHelper.getStartPixel(fzr + 1);
        }

        sb.append(name).append(" .zstop{");
        sb.append("left:").append(lw).append("px;");
        // KEIKAI-215
        // KEIKAI-249
        sb.append('height:').append(fzr > -1 ? toph - 1 : (hidecol ? 0 : toph)).append('px;');
        // sb.append("line-height:").append(toph).append("px;\n");
        sb.append("}");

//        sb.append(name).append(" .zstopi{");
//        sb.append("height:").append(toph).append("px;");
//        sb.append("}");

        sb.append(name).append(" .zstophead{");
        sb.append("height:").append(topheadh).append("px;");
        sb.append("}");

        sb.append(name).append(" .zscornertop{");
        sb.append("left:").append(lw).append("px;");
        sb.append("height:").append(cornertoph).append("px;");
        sb.append("}");

        // relative, so needn't set top position.
        /*
         * sb.append(name).append(" .zstopblock{\n");
         * sb.append("top:").append(topblocktop).append("px;\n");
         * sb.append("}\n");
         */

        // zcss.setRule(name+" .zstopcell",["padding","height","width","line-height"],["0px "+cp+"px 0px "+cp+"px",th+"px",cellwidth+"px",lh+"px"],true,sid);
        sb.append(name).append(" .zstopcell{");
        sb.append("padding:").append("0px ").append(cp).append("px 0px ")
            .append(cp).append("px;");
        sb.append("height:").append(topheadh).append("px;");
        sb.append("width:").append(cellwidth).append("px;");
        sb.append("line-height:").append(topheadh).append("px;");
        sb.append("}");

        // zcss.setRule(name+" .zstopcelltxt","width", celltextwidth
        // +"px",true,sid);
        sb.append(name).append(" .zstopcelltxt{");
        sb.append("width:").append(celltextwidth).append("px;");
        sb.append("}");

        // zcss.setRule(name+" .zsleft",["top","width"],[th+"px",(lw-2)+"px"],true,sid);

        let leftw = lw;
        let leftheadw = leftw;
        let leftblockleft = leftw;

        if (fzc > -1) {
            leftw = leftw + colHelper.getStartPixel(fzc + 1);
        }

        sb.append(name).append(" .zsleft{");
        //ZSS-948: + 1 to avoid 1st row's top border covered by heading's bottom border
        //		sb.append("top:").append(th).append("px;");
        // KEIKAI-215
        sb.append("top:").append(th).append('px;');
//        sb.append('width:').append(fzc > -1 ? leftw: (hiderow ? 0 : leftw + 1)).append('px;');
        // KEIKAI-249
        sb.append('width:').append(fzc > -1 ? leftw - 1: (hiderow ? 0 : leftw)).append('px;');
        sb.append("}");

        sb.append(name).append(" .zslefti{");
        sb.append("width:").append(leftw).append("px;");
        sb.append("}");

        sb.append(name).append(" .zslefthead{");
        // KEIKAI-195
        if (leftheadw === 0) {
            sb.append("opacity:0");
        } else {
            sb.append("width:").append(leftheadw).append("px;");
        }
        sb.append("}");

        sb.append(name).append(" .zsleftblock{");
        sb.append("left:").append(leftblockleft).append("px;");
        sb.append("}");

        sb.append(name).append(" .zscornerleft{");
        sb.append("top:").append(th).append("px;");
        sb.append("width:").append(leftheadw).append("px;");
        sb.append("}");

        // zcss.setRule(name+" .zsleftcell",["height","line-height"],[(rh-1)+"px",(rh)+"px"],true,sid);//for
        // middle the text, i use row leight instead of lh
        sb.append(name).append(" .zsleftcell{");
        sb.append("height:").append(rh).append("px;");
        sb.append("line-height:").append(rh).append("px;");
        sb.append("}");

        // zcss.setRule(name+" .zscorner",["width","height"],[(lw-2)+"px",(th-2)+"px"],true,sid);

        sb.append(name).append(" .zscorner{");
        // KEIKAI-195
        sb.append('width:').append(fzc > -1 ? leftw - 1 : leftw).append('px;');
        sb.append('height:').append(fzr > -1 ? toph - 1 : toph).append('px;');
        sb.append("}");

        sb.append(name).append(" .zscorneri{");
        sb.append("width:").append(lw - 1).append("px;");
        sb.append("height:").append(th - 1).append("px;");
        sb.append("}");

        sb.append(name).append(" .zscornerblock{");
        sb.append("left:").append(lw).append("px;");
        sb.append("top:").append(th).append("px;");
        sb.append("}");

        // zcss.setRule(name+" .zshboun","height",th+"px",true,sid);
        //		sb.append(name).append(" .zshboun{");
        //		sb.append("height:").append(th).append("px;");
        //		sb.append("}");

        // zcss.setRule(name+" .zshboun","height",th+"px",true,sid);
        /*		sb.append(name).append(" .zshboun{\n");
                sb.append("height:").append(th).append("px;\n");
                sb.append("}\n");
        */
        // zcss.setRule(name+" .zshbouni","height",th+"px",true,sid);
        sb.append(name).append(" .zshbouni{");
        sb.append("height:").append(th).append("px;");
        sb.append("}");

        sb.append(name).append(" .zsfztop{");
        sb.append("border-bottom-style:").append(fzr > -1 ? "solid" : "none").append(";");
        sb.append("}");
        sb.append(name).append(" .zsfztop .zswidgetpanel{");
        sb.append("left:").append(-lw).append("px;");
        sb.append("}");

        sb.append(name).append(" .zsfzcorner{");
        sb.append("border-bottom-style:").append(fzr > -1 ? "solid" : "none").append(";");
        sb.append("}");
        sb.append(name).append(" .zsfzcorner .zswidgetpanel{");
        sb.append("left:1px;");
        sb.append("top:1px;");
        sb.append("}");

        sb.append(name).append(" .zsfzleft{");
        sb.append("border-right-style:").append(fzc > -1 ? "solid" : "none").append(";");
        sb.append("}");
        sb.append(name).append(" .zsfzleft .zswidgetpanel{");
        sb.append("top:").append(-th).append("px;");
        sb.append("}");

        sb.append(name).append(" .zsfzcorner{");
        sb.append("border-right-style:").append(fzc > -1 ? "solid" : "none")
            .append(";");
        sb.append("}");
    },
    _prepareColHeaderStyleSheet: function (sb) {
        let _spreadsheet = this._sheet._wgt;
        let name = "#" + _spreadsheet.uuid;
        let colHelper = this._sheet.custColWidth;
        let cp = _spreadsheet.getCellPadding();
        if (colHelper._hasCustom()) {
            colHelper.custom.forEach(function (info) {
                let hidden = info[3];
                let width = hidden ? 0 : info[1];
                let cid = info[2];

                let celltextwidth = width - 2 * cp;

                // bug 1989680
                if (celltextwidth < 0)
                    celltextwidth = 0;

                let cellwidth = width;

                // KEIKAI-62: no need for hidden column
                if (width > 0) {
                    sb.append(name).append(" .zsw").append(cid).append("{");
                    sb.append("width:").append(cellwidth).append("px;");
                    sb.append("}");

                    sb.append(name).append(" .zswi").append(cid).append("{");
                    sb.append("width:").append(celltextwidth).append("px;");
                    sb.append("}");
                }
            });
        }
    },
    _prepareRowHeaderStyleSheet: function (sb) {
        let _spreadsheet = this._sheet._wgt;
        let name = "#" + _spreadsheet.uuid;
        let rowHelper = this._sheet.custRowHeight;
        let cp = _spreadsheet.getCellPadding();
        if (rowHelper._hasCustom()) {
            rowHelper.custom.forEach(function (info) {
                let hidden = info[3];
                let height = hidden ? 0 : info[1];
                let cid = info[2];
                let cellheight = height;

                // KEIKAI-62: no need for hidden row
                if (height > 0) {
                    sb.append(name).append(" .zsh").append(cid).append("{");
                    sb.append("height:").append(height).append("px;");
                    sb.append("}");

                    sb.append(name).append(" .zshi").append(cid).append("{");
                    sb.append("height:").append(cellheight).append("px;");
                    sb.append("border-bottom-width:1px;");
                    sb.append("}");

                    sb.append(name).append(" .zslh").append(cid).append("{");
                    sb.append("height:").append(height).append("px;");
                    sb.append("line-height:").append(height).append("px;");
                    sb.append("border-bottom-width:1px;");
                    sb.append("}");

                    sb.append(name).append(" .zshr").append(cid).append("{");
                    sb.append("max-height:").append(height).append("px;");
                    sb.append("}");
                }
            });
        }
        //TODO: seems no need
        sb.append(".zs_header{}");// for indicating add new rule before this
    },
    _prepareMergeStyleSheet: function (sb) {
        let _spreadsheet = this._sheet._wgt;
        let name = "#" + _spreadsheet.uuid;

        let colHelper = this._sheet.custColWidth;
        let rowHelper = this._sheet.custRowHeight;
        let mmhelper = this._sheet.mergeMatrix;

        let cp = _spreadsheet.getCellPadding();
        let ranges = mmhelper.getMergeMatrix();
        let defaultSize = colHelper.size;
        let defaultRowSize = rowHelper.size;
        if (ranges && ranges.length) {
            ranges.forEach(function (block) {
                let left = block.left;
                let right = block.right;
                let width = 0;
                for (let i = left; i <= right; i++) {
                    let info = colHelper.getMeta(i);
                    if (info != null) {
                        let hidden = info[3];
                        let colSize = hidden ? 0 : info[1];
                        width += colSize;
                    } else {
                        width += defaultSize ;
                    }
                }
                let top = block.top;
                let bottom = block.bottom;
                let height = 0;
                for (let i = top; i <= bottom; i++) {
                    let info = rowHelper.getMeta(i);
                    if (info != null) {
                        let hidden = info[3];
                        let rowSize = hidden ? 0 : info[1];
                        height += rowSize;
                    } else {
                        height += defaultRowSize ;
                    }
                }

                if (width <= 0 || height <= 0) { //total hidden
                    sb.append(name).append(" .zsmerge").append(block.id).append("{");
                    sb.append("display:none;");
                    sb.append("}");

                    sb.append(name).append(" .zsmerge").append(block.id);
                    sb.append(" .zscelltxt").append("{");
                    sb.append("display:none;");
                    sb.append("}");
                } else {
                    let celltextwidth = width - 2 * cp;
                    let celltextheight = height;

                    let cellwidth = width;
                    let cellheight = height;

                    sb.append(name).append(" .zsmerge").append(block.id).append("{");
                    sb.append("width:").append(cellwidth).append("px;");
                    sb.append("height:").append(cellheight).append("px;");
                    sb.append("display:inline-block;"); // ZSS-330, ZSS-382:  the left-top cell must display
                    sb.append("border-bottom-width: 1px;"); // re-apply bottom border for grid line; Or grid line will be missed if row was hidden
                    sb.append("}");

                    sb.append(name).append(" .zsmerge").append(block.id);
                    sb.append(" .zscelltxt").append("{");
                    sb.append("width:").append(celltextwidth).append("px;");
                    sb.append("height:").append(celltextheight).append("px;");
                    sb.append("}");
                }
            });
        }
    },
    _prepareOtherStyleSheet: function (sb) {
        let _spreadsheet = this._sheet._wgt;
        let name = "#" + _spreadsheet.uuid;
        sb.append(name).append(" .zs_indicator_")
            .append(_spreadsheet.getSheetId()).append("{}");// for indicating the css is load ready

        if (!zk.ie && !zk.mobile) {
            const wgt = this._sheet._wgt;
            // KEIKAI-198
            const top = wgt.isColumnHeadHidden() ? 2 : 22;
            sb.append(name).append(' .simplebar-track.simplebar-vertical .simplebar-scrollbar::before{');
            sb.append('top:').append(top).append('px;');
            sb.append('bottom:17px;');
            sb.append('}');

            const left = wgt.isRowHeadHidden() ? 2 : 37;
            sb.append(name).append(' .simplebar-track.simplebar-horizontal .simplebar-scrollbar::before{');
            sb.append('left:').append(left).append('px;');
            sb.append('right:17px;');
            sb.append('}');
        }
    }
});

})();