/* Range.java

{{IS_NOTE
	Purpose:
		
	Description:
		
	History:
		2013/5/1 , Created by dennis
}}IS_NOTE

Copyright (C) 2013 Potix Corporation. All Rights Reserved.

{{IS_RIGHT
}}IS_RIGHT
*/
package io.keikai.api;

import java.util.List;
import java.util.Set;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.stream.Stream;

import io.keikai.api.model.Book;
import io.keikai.api.model.CellData;
import io.keikai.api.model.CellStyle;
import io.keikai.api.model.Chart;
import io.keikai.api.model.Color;
import io.keikai.api.model.EditableCellStyle;
import io.keikai.api.model.EditableFont;
import io.keikai.api.model.Font;
import io.keikai.api.model.Hyperlink;
import io.keikai.api.model.Picture;
import io.keikai.api.model.Sheet;
import io.keikai.api.model.SheetProtection;
import io.keikai.api.model.Validation;
import io.keikai.model.CellRegion;
import io.keikai.range.SRange;

/**
 * Range can represent a cell, a row, a column, a selection of cells containing one or 
 * more contiguous blocks of cells, or a 3-D blocks of cells. <br/>
 * You have to use this class's API to do any operation of the {@link Sheet}, then the upload will sync to the UI automatically.<br/>
 * To get the instance of a {@link Range}, please use the {@link Ranges} API.
 * 
 * <br/>
 * Note : the range API doesn't check the sheet protection, if you care it, you have to check it by calling {@link #isProtected()} before you do any operation.
 * 
 * @author dennis
 * @see Ranges
 * @since 3.0.0
 */
public interface Range {

	public enum PasteType{
		ALL,
		ALL_EXCEPT_BORDERS,
		COLUMN_WIDTHS,
		COMMENTS,
		FORMATS/*all formats*/,
		FORMULAS/*include values and formulas*/,
		FORMULAS_AND_NUMBER_FORMATS,
		VALIDATAION,
		VALUES,
		VALUES_AND_NUMBER_FORMATS;
	}
	
	public enum PasteOperation{
		ADD,
		SUB,
		MUL,
		DIV,
		NONE;
	}
	
	public enum ApplyBorderType{
		FULL,
		EDGE_BOTTOM,
		EDGE_RIGHT,
		EDGE_TOP,
		EDGE_LEFT,
		OUTLINE,
		INSIDE,
		INSIDE_HORIZONTAL,
		INSIDE_VERTICAL,
		DIAGONAL,
		DIAGONAL_DOWN,
		DIAGONAL_UP
	}
	
	/** Shift direction of insert api**/
	public enum InsertShift{
		DEFAULT,
		RIGHT,
		DOWN
	}
	/** 
	 * Copy origin format/style of insert
	 **/
	public enum InsertCopyOrigin{
		FORMAT_NONE,
		FORMAT_LEFT_ABOVE,
		FORMAT_RIGHT_BELOW,
	}
	/** Shift direction of delete api**/
	public enum DeleteShift{
		DEFAULT,
		LEFT,
		UP
	}
	
	public enum SortDataOption{
		NORMAL_DEFAULT,
		TEXT_AS_NUMBERS
	}
	
	public enum AutoFilterOperation{
		AND,
		BOTTOM10,
		BOTTOM10PERCENT,
		OR,
		TOP10,
		TOP10PERCENT,
		VALUES
	}
	
	public enum AutoFillType{
		COPY,
		DAYS,
		DEFAULT,
		FORMATS,
		MONTHS,
		SERIES,
		VALUES,
		WEEKDAYS,
		YEARS,
		GROWTH_TREND,
		LINER_TREND
	}
	
	public enum SheetVisible {
		VISIBLE,
		HIDDEN,
		VERY_HIDDEN,
	}
	
	//ZSS-939: the sequence must be consistent with io.keikai.model.impl.CellAttribute
	//@since 3.8.0
	public enum CellAttribute {
		ALL, TEXT, STYLE, SIZE, MERGE, COMMENT;
	}
	
	public ReadWriteLock getLock();
	
	/**
	 * Gets the book of this range
	 * @return book
	 */
	public Book getBook();
	
	/**
	 * Gets the sheet of this range
	 * @return sheet
	 */
	public Sheet getSheet();
	
	/**
	 * Gets the left column of this range
	 * @return the left column
	 */
	public int getColumn();
	
	/**
	 * Gets the top row of this range
	 * @return the top row
	 */
	public int getRow();
	
	/**
	 * Gets the right/last column of this range
	 * @return
	 */
	public int getLastColumn();
	
	/**
	 * Gets the bottom/last row of this range
	 * @return
	 */
	public int getLastRow();
	
	
	/**
	 * Gets the row count of this range
	 * @return count of row of this range
	 */
	public int getRowCount();
	
	/**
	 * Gets the column count of this range
	 * @return count of column of this range
	 */
	public int getColumnCount();
	
	/**
	 * Gets cell-style-helper, this helper helps you to create new style, font or color 
	 * @return
	 */
	public CellStyleHelper getCellStyleHelper();
	
	/**
	 * Runs runer under synchronization protection
	 * @param run the runner
	 */
	public void sync(RangeRunner run);
	/**
	 * Visits all cells in this range with synchronization protection, make sure you call this in a limited range, 
	 * don't use it for all row/column selection, it will spend much time to iterate the cell 
	 * @param visitor the cell visitor 
	 */
	public void visit(final CellVisitor visitor);

	/**
	 * Return a new range by shifting the current Range according to the offsets but still keep the original height and width e.g.
	 * <pre>
	 * {@code
	 * // range is A1:B2
	 * range.toShiftedRange(1, 0); //returns A2:B3
	 * }
	 * </pre>
	 * @param rowOffset row offset of the new range. Positive number shifts it down, negative number shifts it up.
	 * @param colOffset column offset of the new range. Positive number shifts it right, negative number shifts it left.
	 * @return the new shifted range
	 */
	public Range toShiftedRange(int rowOffset,int colOffset);
	
	/**
	 * Returns a new range having one cell according to the offset based on the left-top corner. For example:
	 * <pre>
	 * {@code
	 * //range is A1:B2
	 * range.toCellRange(0, 0); // returns A1
	 * }
	 * </pre>
	 * @param rowOffset row offset of the cell.
	 * @param colOffset column offset of the cell.
	 * @return the new range of the cell
	 */
	public Range toCellRange(int rowOffset,int colOffset);
	
	/**
	 *  Return a range that represents all columns and between the first-row and last-row of this range.
	 *  It is a useful when you want to manipulate entire row (such as deleting rows)
	 **/
	public Range toRowRange();
	
	/**
	 *  Return a range that represents all rows and between the first-column and last-column of this range
	 *  It is a useful when you want to manipulate entire column (such as deleting columns)
	 **/
	public Range toColumnRange();
	
	/**
	 * Check if this range represents a whole column, which means all rows are included,
	 */
	public boolean isWholeColumn();
	/**
	 * Check if this range represents a whole row, which means all column are included,
	 */
	public boolean isWholeRow();
	/**
	 * Check if this range represents a whole sheet, which mean all column and row are included, 
	 */
	public boolean isWholeSheet();
	
	/*
	 * ==================================================
	 * operations of cell related API
	 * ==================================================
	 */
	
	/**
	 * Clears contents
	 */
	public void clearContents();

	/**
	 * Clears styles
	 */
	public void clearStyles();
	
	/**
	 * Clears contents, styles and unmerge this range.
	 * @Since 3.5.0
	 */
	public void clearAll();

	/**
	 * Pastes to destination. Support pasting cells among 2 books.
	 * @param dest the destination. It can be a range of a different book.
	 * @return a Range contains the final pasted range. paste to a protected sheet will always cause paste return null.
	 */
	public Range paste(Range dest);
	
	/**
	 * Cut and paste to destination. Support pasting cells among 2 books.
	 */
	public Range paste(Range dstRange, boolean cut);	
	
	/**
	 * Pastes to destination. Support pasting cells among 2 books.
	 * @param dest the destination. It can be a range of a different book.
	 * @param type the paste type
	 * @param op the paste operation
	 * @param skipBlanks skip blanks or not
	 * @param transpose transpose the cell or not
	 * @return a Range contains the final pasted range. paste to a protected sheet will always cause paste return null.
	 * @throws IllegalOpArgumentException 
	 */
	public Range pasteSpecial(Range dest,PasteType type,PasteOperation op,boolean skipBlanks,boolean transpose);
	
	/**
	 * apply borders
	 * @param applyType the apply type
	 * @param borderType the border type
	 * @param htmlColor the color (#rgb-hex-code, e.x #FF00FF)
	 */
	public void applyBorders(ApplyBorderType applyType,
			CellStyle.BorderType borderType,String htmlColor);

	/**
	 * @return true if any merged cell inside (fully contains or overlaps) this range
	 */
	public boolean hasMergedCell();
	
	/**
	 * @return true if entire range is a merged cell.
	 */
	public boolean isMergedCell();
	
	/**
	 * Merges the range
	 * @param across true if merge horizontally only
	 */
	public void merge(boolean across);
	
	/**
	 * Unmerge the range
	 */
	public void unmerge();
	
	/**
	 * Insert new cells to the area of this range.<br/> 
	 * To insert a row, you have to call {@link Range#toRowRange()} first, to insert a column, you have to call {@link Range#toColumnRange()} first.
	 * @param shift the shift direction of original cells
	 * @param copyOrigin copy the format from nearby cells when inserting new cells 
	 */
	public void insert(InsertShift shift,InsertCopyOrigin copyOrigin);
	
	/**
	 * Delete cells of the range. <br/>
	 * To delete a row, you have to call {@link Range#toRowRange()} first, to delete a column, you have to call {@link Range#toColumnRange()} first.
	 * @param shift the shift direction when deleting.
	 */
	public void delete(DeleteShift shift);
	
	/**
	 * Sort range 
	 * @param desc true for descent, false for ascent
	 */
	public void sort(boolean desc);
	
	/**
	 * Sort range
	 * @param desc true for descent, false for ascent
	 * @param header includes header or not
	 * @param matchCase matches character chase of not
	 * @param sortByRows sorts by row or not
	 * @param dataOption data option for sort
	 */
	public void sort(boolean desc,
			boolean header, 
			boolean matchCase, 
			boolean sortByRows, 
			SortDataOption dataOption);
	
	/**
	 * Sort range
	 * @param index1 the sort index 1
	 * @param desc1 true for descent, false for ascent of index 1
	 * @param header includes header or not
	 * @param matchCase matches character chase of not
	 * @param sortByRows  sorts by row or not
	 * @param dataOption1 data option 1 for sort
	 * @param index2 the sort index 2
	 * @param desc2 true for descent, false for ascent of index 2
	 * @param dataOption2 data option 2 for sort
	 * @param index3 the sort index 3
	 * @param desc3 true for descent, false for ascent of index 3
	 * @param dataOption3 data option31 for sort
	 */
	public void sort(Range index1,boolean desc1,SortDataOption dataOption1,
			Range index2,boolean desc2,SortDataOption dataOption2,
			Range index3,boolean desc3,SortDataOption dataOption3,
			boolean header,
			/*int orderCustom, //not implement*/
			boolean matchCase, 
			boolean sortByRows 
			/*int sortMethod, //not implement*/
			);
	
	/**
	 * According to current range, fills data to destination range automatically
	 * @param dest the destination range
	 * @param fillType the fill type
	 */
	public void autoFill(Range dest,AutoFillType fillType);
	
	/**
	 * Fills cells by copying from first/top row data
	 */
	public void fillDown();
	
	/**
	 * Fills cells by copying from last/right column data
	 */
	public void fillLeft();
	
	/**
	 * Fills cells by copying from bottom row data
	 */
	public void fillUp();
	
	/**
	 * Fills cells by copying from first/left column data
	 */
	public void fillRight();
	
	/** Shifts/moves cells with a offset row and column**/
	public void shift(int rowOffset,int colOffset);
	
	/**
	 * Sets the width(in pixel) of column in this range, it effect to whole column. 
	 * @param widthPx width in pixel
	 * @see #toColumnRange()
	 */
	public void setColumnWidth(int widthPx);
	/**
	 * Sets the height(in pixel) of row in this range, it effect to whole row.
	 * @param heightPx height in pixel
	 * @see #toRowRange()
	 */
	public void setRowHeight(int heightPx);
	
	/**
	 * Sets the height(in pixel) of row in this range and specify it's custom size or not.
	 * @param heightPx height in pixel
	 * @param isCustom true if it's set by users manually, false if it's determined by the system automatically
	 * @since 3.0.1
	 */
	public void setRowHeight(int heightPx, boolean isCustom);
	/* 
	 * ==================================================
	 * cell related API
	 * ==================================================
	 */
	/**
	 * Sets cell style, applies it to all cells of this range 
	 * @param nstyle new cell style
	 * @see #getCellStyleHelper()
	 * @see #getCellStyle()
	 */
	public void setCellStyle(CellStyle nstyle);

	/**
	 * Works like a user entering a text in a cell via a browser. So, this method supports "Smart Input", and it will parse the text and turn the text into a String, Number, Date, or Boolean according to the supported input format pattern. To enter a formula, the text should stat with <code>=</code>. <br/>
	 * This method applies all cells of this range.
	 * @param editText the text to enter, it can be a string, integer string, date string or a formula (start with '=')
	 * @throws IllegalFormulaException
	 */
	public void setCellEditText(String editText);

	/**
	 * Works like a user entering an array formula in a cell or a range via a browser.
	 * So, this method supports "Smart Input", and it will parse the text and turn the text into a String, Number, Date, or Boolean according to the supported input format pattern. To enter a formula, the text should stat with <code>=</code>. <br/>
	 * This method applies all cells of this range.
	 * @param editText the text to enter, it can be a string, integer string, date string or a formula (start with '=')
	 * @throws IllegalFormulaException
	 * @since 6.0.0
	 */
	public void setArrayFormula(String editText);

	/**
	 * Sets cell data value and type. Notice that it doesn't parse the specified String object. If you call:
	 * {@code setCellValue("100") }, it will produce a text cell instead of an integer, one hundred. If you need a number cell, you need to call {@code setCellValue(100) }. The same rule applies to Date and Boolean.<br/>
	 * This method applies all cells of this range.
	 * @param value the cell value, could be null, String, Number, Date or Boolean
	 * @see #setCellEditText(String)
	 */
	public void setCellValue(Object value);

	/**
	 * Sets cell data values, applies them to all cells from left to right and top to bottom consistently.
	 * For example:
	 * <h2>Set values in a row</h2>
	 * <pre>{@code
	 * Range rA1C1 = Ranges.range("A1:C1");
	 * rA1C1.setCellValues(3,2,1);
	 * }</pre>
	 * The result is [3,2,1] from A1 to C1
	 *
	 * <h2>Set values in a column</h2>
	 * <pre>{@code
	 * Range rA1A3 = Ranges.range("A1:A3");
	 * rA1A3.setCellValues(1,2,3);
	 * }</pre>
	 * The result is [1,2,3] from A1 to A3<br/>
	 *
	 * <h2>Set values in an area</h2>
	 * <pre>{@code
	 * Range rA1C3 = Ranges.range("A1:C3");
	 * rA1C3.setCellValues(9, 8, 7, 6, 5, 4, 3,2,1);
	 * }</pre>
	 * The result is [9, 8, 7, 6, 5, 4, 3, 2, 1] from A1, B1, C1, A2, ... to C3<br>
	 * <h2>Cell and values count mismatch</h2>
	 * If you pass more values than cell number, then it just sets cells in this range, not over-fill.
	 * <pre>{@code
	 * rA1C1.setCellValues(1, 2, 3, 4, 5);
	 * }</pre>
	 * The result is [1,2,3] from A1 to C1.<br/><br/>
	 * If you pass fewer values than cell number, then it just set cells according the number of values.
	 * <pre>{@code
	 * rA1C1.setCellValues(1, 2);
	 * }</pre>
	 * The result is [1,2] from A1 to B1.<br/><br/>
	 * <h2>Transpose 2D Array when necessary</h2>
	 * Since this method sets cells row by row, if you have a 2D array that contains column by column data.
	 * Then you can transpose the array first and pass to the method.
	 * @param values all cell values, could be null, String, Number, Date or Boolean
	 * @since 5.5.0
	 */
	public void setCellValues(Object... values);
	
	/**
	 * Sets cell hyperlink, applies it too all cells
	 * @param type the hyperlink type
	 * @param address the address, e.x http://www.zkoss.org
	 * @param label the label to display
	 */
	public void setCellHyperlink(Hyperlink.HyperlinkType type,String address,String label);
	
	
	/**
	 * Gets the first cell(top-left) {@link Hyperlink} object of this range.
	 * @return
	 */
	public Hyperlink getCellHyperlink();
	
	/**
	 * Gets the first cell(top-left) style of this range
	 * 
	 * @return cell-style of this cell, if this cell doesn't has cell-style, it will check row/column's cell-style and then sheet's default cell-style 
	 */
	public CellStyle getCellStyle();
	
	/**
	 * Gets the first cell(top-left) data of this range
	 * @return
	 */
	public CellData getCellData();
	
	/**
	 * Gets the first cell(top-left) edit text of this range
	 * @return edit text
	 * @see CellData#getEditText()
	 */
	public String getCellEditText();
	
	/**
	 * Gets the first cell(top-left) formatted text of this range
	 * @return format text
	 * @see CellData#getFormatText()
	 */
	public String getCellFormatText();
	
	/**
	 * Gets the first cell(top-left)  format of this range
	 * @return
	 * @since 3.5.0
	 */
	public String getCellDataFormat();
	
	/**
	 * Gets the first cell(top-left) value of this range
	 * @return value object or null for a blank cell
	 * @see CellData#getValue()
	 */
	public Object getCellValue();

	/**
	 * Gets all cell values from left to right and top to bottom of this range in
	 * one dimensional list. Blank cell will be null value instead.<br>
	 * If the range is A1:C3, the method will return you a list of cell values like:<br>
	 * <code>A1, B1, C1, A2, B2, C2, A3, B3, C3</code>
	 * @return values a object list
	 * @since 5.5.0
	 */
	public List<Object> getCellValues();
	
	/* 
	 * ==================================================
	 * sheet relative API
	 * ==================================================
	 */
	/**
	 * Displays sheet grid-lines or not
	 * @param enable true to display
	 */
	public void setDisplaySheetGridlines(boolean enable);
	
	/**
	 * @return true if display sheet grid-lines is enabled
	 */
	public boolean isDisplaySheetGridlines();
	
	/**
	 * Hide or unhide rows or columns.<br/> 
	 * To hide/unhide a row, you have to call {@link Range#toRowRange()} first, to hide/un-hide a column, you have to call {@link Range#toColumnRange()} 
	 * or a whole column range. 
	 * @param hidden hide or not
	 */
	public void setHidden(boolean hidden);
	
	/**
	 * Sets the sheet name
	 * @param name new sheet name, it must be not same as another sheet name in it's owner book.
	 */
	public void setSheetName(String name);
	
	/**
	 * Gets the sheet name
	 * @return sheet name
	 */
	public String getSheetName();
	
	/**
	 * Sets the sheet order
	 * @param pos the position
	 */
	public void setSheetOrder(int pos);
	
	/**
	 * Gets the sheet order
	 * @return
	 */
	public int getSheetOrder();
	
	/**
	 * @return true if the specified range is protected; i.e. sheet is protected
	 * and some cells in the range are locked
	 */
	public boolean isProtected();
	
	/**
	 * @return true if auto filter is enabled.
	 */
	public boolean isAutoFilterEnabled();
	
	/**
	 * To find a range of cells for applying auto filter according to this range.
	 * Usually, these two ranges are different.
	 * This method searches the filtering range through a specific rules. 
	 * @return a range of cells for applying auto filter or null if can't find one from this Range. 
	 */
	// Refer to ZSS-246.
	public Range findAutoFilterRange();
	
	/**
	 * Enable/disable autofilter of the sheet
	 * @param enable true to enable
	 **/
	public void enableAutoFilter(boolean enable);

	/**
	 * Enables auto filter and set extra condition
	 * @param field the filed index (according to current range, 1 base)
	 * @param filterOp auto filter operation
	 * @param criteria1 criteria for autofilter
	 * @param criteria2 criteria for autofilter
	 * @param showButton true/false for show/hide dropdown button, null will keep the original setting.
	 */
	public void enableAutoFilter(int field, AutoFilterOperation filterOp, Object criteria1, Object criteria2, Boolean showButton);
	
	/**
	 * Clears condition of filter, show all the data
	 **/
	public void resetAutoFilter();
	
	/** 
	 * Re-applies the filter, filter by last condition and data again. Call this if the data was change and you want to re-new the filter result.
	 **/
	public void applyAutoFilter();
	
	/**
	 * Adds picture to sheet
	 * @param anchor the anchor for picture
	 * @param image the image binary array
	 * @param format the image format
	 * @return the new added picture
	 */
	public Picture addPicture(SheetAnchor anchor,byte[] image,
			Picture.Format format);
	
	/**
	 * Deletes picture that in sheet
	 * @param picture
	 */
	public void deletePicture(Picture picture);
	
	/**
	 * Moves picture
	 * @param anchor the anchor to re-allocate
	 * @param picture the picture to re-allocate
	 */
	public void movePicture(SheetAnchor anchor,Picture picture);

	/**
	 * Updates picture with anchor type
	 * @since 5.12.0
	 */
	public void updatePicture(Picture picture, SheetAnchor.AnchorType anchorType);

	/**
	 * Adds chart to sheet
	 * @param anchor the destination anchor of the chart
	 * @param type the chart type
	 * @param grouping the chart grouping
	 * @param pos the legend position
	 * @return the new added chart
	 */
	//currently, we only support to modify chart in XSSF
	public Chart addChart(SheetAnchor anchor,
			Chart.Type type, Chart.Grouping grouping, Chart.LegendPosition pos);
	
	/**
	 * Deletes chart
	 * @param chart the chart to delete
	 */
	//currently, we only support to modify chart in XSSF
	public void deleteChart(Chart chart);
	
	/**
	 * Moves chart to new location
	 * @param anchor the new location to move
	 * @param chart the chart to move
	 */
	//currently, we only support to modify chart in XSSF
	public void moveChart(SheetAnchor anchor,Chart chart);
	
	/**
	 * Creates a new sheet and appends it as the last sheet. To move the new sheet to the specific position, see {@link #setSheetOrder(int)}.
	 * @param name the sheet name, it must not be same as another sheet name in book of this range. Otherwise, it takes no effect.
	 * @return the new created sheet
	 */
	public Sheet createSheet(String name);
	
	
	/**
	 * Clone this sheet; create a sheet and copy the contents of this sheet; then add it as the last sheet.
	 * @param name the sheet name, it must not be the same as another sheet name in book of this range
	 * @return the new created sheet
	 * @since 3.6.0
	 */
	public Sheet cloneSheet(String name);
	
	/**
	 * Set the freeze panel
	 * @param rowfreeze the number of row to freeze, 0 means no freeze
	 * @param columnfreeze the number of column to freeze, 0 means no freeze
	 */
	public void setFreezePanel(int rowfreeze,int columnfreeze);

	
	/**
	 * Deletes sheet.
	 * Note: You couldn't delete last sheet of a book.
	 */
	public void deleteSheet();


	/**
	 * Notify this range has been changed, then spreadsheet will render the changed cells to a browser. It's usually used with {@link #setAutoRefresh(boolean)}.
	 */
	public void notifyChange();
	
	
	/**
	 * Notify the whole book of specified variables change. Then spreadsheet will re-evaluate those cells that reference to these variables.
	 * @param variables changed variables 
	 */
	public void notifyChange(String[] variables);
	
	
	
	/**
	 * get formatted string of this range
	 * @return
	 */
	public String asString();
	
	/**
	 * a cell style helper to create style relative object for cell
	 * @author dennis
	 */
	public interface CellStyleHelper {

		/**
		 * create a new cell style builder and clone attribute from src if it is not null
		 * @param src the source to clone, could be null
		 * @return the new cell style
		 * @since 5.3.0
		 */
		CellStyle.Builder builder(CellStyle src);


		/**
		 * create a new cell style and clone attribute from src if it is not null
		 * @param src the source to clone, could be null
		 * @return the new cell style
		 * @deprecated As if release 5.3.0, replaced by {@link #builder(CellStyle)} to create an immutable
		 * cell style
		 */
		@Deprecated public EditableCellStyle createCellStyle(CellStyle src);

		/**
		 * create a new font and clone attribute from src if it is not null
		 * @param src the source to clone, could be null
		 * @return the new font
		 * @deprecated As if release 5.3.0, replaced by {@link #builder(CellStyle)} to create an immutable font
		 */
		@Deprecated public EditableFont createFont(Font src);
		
		/**
		 * create a color object from a htmlColor expression
		 * @param htmlColor html color expression, ex. #FF00FF
		 * @return a Color object
		 */
		public Color createColorFromHtmlColor(String htmlColor);
		
		/**
		 * find the font with given condition
		 * @param boldweight
		 * @param color
		 * @param fontHeight
		 * @param fontName
		 * @param italic
		 * @param strikeout
		 * @param typeOffset
		 * @param underline
		 * @return null if not found
		 * @deprecated As if release 5.3.0, replaced by {@link CellStyleHelper#builder(CellStyle)} to create an immutable font
		 */
		@Deprecated public Font findFont(Font.Boldweight boldweight, Color color,
				int fontHeight, String fontName, boolean italic,
				boolean strikeout, Font.TypeOffset typeOffset, Font.Underline underline);

		/**
		 * Check if this style still available
		 */
		public boolean isAvailable(CellStyle style);
	}

	/**
	 * Notify the component that a chart has change, e.g. call it after chart data changes.
	 * @param chart the chart that contains change
	 * @since 3.5.0
	 */
	public void updateChart(Chart chart);

	/**
	 * Notify the component that a chart has change, e.g. call it after chart data changes.
	 * @param chart the chart that contains change
	 * @param anchorType
	 * @since 5.12.0
	 */
	public void updateChart(Chart chart, SheetAnchor.AnchorType anchorType);
	
	
	/**
	 * Get internal range implementation
	 * @since 3.5.0
	 */
	public SRange getInternalRange();
	
	/**
	 * Create a name that refers to this range. You can get this named ranage by {@link Ranges#rangeByName(Sheet, String)}.
	 * @param nameName name of the range that you can refer in formulas.
	 * @since 3.5.0
	 */
	public void createName(String nameName);
	
	/**
	 * @return true if the sheet is protected
	 * @since 3.5.0
	 */
	public boolean isSheetProtected();
	
	/**
	 * Protect a {@link Sheet} so that it cannot be modified. You can call 
	 * this method on a protected sheet to change its allowed options if the 
	 * same password is supplied; otherwise, it is ignored. A protected sheet
	 * can be unprotected by calling {@link #unprotectSheet(String password)}.
	 * 
	 * change the protection options; make sure provide 
	 * @param password a case-sensitive password for the sheet; null or empty string means protect the sheet without password.
	 * @param allowSelectingLockedCells true to allow select locked cells; default to true.
	 * @param allowSelectingUnlockedCells true to allow select unlocked cells; default to true.
	 * @param allowFormattingCells true to allow user to format any cell on the protected sheet; default false.
	 * @param allowFormattingColumns true to allow user to format any columns on the protected sheet; default false.
	 * @param allowFormattingRows true to allow user to format any rows on the protected sheet; default false.
	 * @param allowInsertColumns true to allow user to insert columns on the protected sheet; default false.
	 * @param allowInsertRows true to allow user to insert rows on the protected sheet; default false.
	 * @param allowInsertingHyperlinks true to allow user to insert hyperlinks on the protected sheet; default false.
	 * @param allowDeletingColumns true to allow user to delete columns on the protected sheet; default false.
	 * @param allowDeletingRows true to allow user to delete rows on the protected sheet; default false.
	 * @param allowSorting true to allow user to sort on the protected sheet; default false.
	 * @param allowFiltering true to allow user to set filters on the protected sheet; default false.
	 * @param allowUsingPivotTables true to allow user to use pivot table reports on the protected sheet; default false.
	 * @param drawingObjects true to protect objects; default to false.
	 * @param scenarios true to protect scenarios; default to true.
	 * 
	 * @see #protectSheet(io.keikai.api.SheetProtection)
	 * @see #unprotectSheet(String password)
	 * @see #isProtected()
	 * @since 3.5.0
	 */
	public void protectSheet(String password,  
			boolean allowSelectingLockedCells, boolean allowSelectingUnlockedCells,  
			boolean allowFormattingCells, boolean allowFormattingColumns, boolean allowFormattingRows, 
			boolean allowInsertColumns, boolean allowInsertRows, boolean allowInsertingHyperlinks,
			boolean allowDeletingColumns, boolean allowDeletingRows, 
			boolean allowSorting, boolean allowFiltering, 
			boolean allowUsingPivotTables, boolean drawingObjects, boolean scenarios);

	/**
	 * Protect a {@link Sheet} so that it cannot be modified.
	 * @param protection
	 * @see io.keikai.api.SheetProtection
	 * @since 5.0.0
	 */
	public void protectSheet(io.keikai.api.SheetProtection protection);

	/**
	 * Removes protection from a sheet. This method has no effect if the sheet 
	 * isn't protected.
	 * @param password a case-sensitive password used to unprotect the sheet. If
	 * the sheet isn't protected with a password, this argument is ignored. If you
	 * omit this argument for a sheet that is protected with a password, you'll
	 * be prompted for the password.
	 * @since 3.5.0 
	 */
	public boolean unprotectSheet(String password);
	
	/**
	 * Gets {@link SheetProtection} which tells what are allowed operations for
	 * a protected sheet of this range.
	 * @return
	 * @since 3.5.0
	 */
	public SheetProtection getSheetProtection();

	/**
	 * validate the user's input.
	 * @param editText
	 * @return the Validation that constraint the input; null if none.
	 */
	public Validation validate(final String editText);

	/**
	 * Add if not exist or modify an existing {@link Validation} to this range.
	 * 
	 * There are two ways to input list items for data validation in argument
	 * "formula1" and "formula2". One is to specify a range via a formula such 
	 * as "=A1:A4" where the values in the cells of the range is the constrained 
	 * value list; another is to input literally the constrained value list
	 * delimited by comma such as "1, 2, 3, 4". Note that when you input a
	 * formula, must lead the formula with an equal sign('='). 
	 *  
	 * @param validationType the type of this validation  
	 * @param ignoreBlank true if blank values are permitted. 
	 * @param operatorType the operator for this validation
	 * @param inCellDropDown true if want to display dropdown list for acceptable values.
	 * @param formula1 the value or expression associated with conditional format or data validation.
	 * @param formula2 the 2nd part of a conditional format or data validation. Useful only when operatorType is BETWEEN or NOT_BETWEEN.
	 * @param showInput true to show the input message.
	 * @param inputTitle title for the data-validation input dialog box.
	 * @param inputMessage message for the data-validation input dialog box.
	 * @param showError true to show the error message.
	 * @param alertStyle validation alert style.
	 * @param errorTitle title of the data validation error dialog.
	 * @param errorMessage data validation error message.
	 * 
	 * @See {@link #getValidations()}
	 * @since 3.5.0
	 */
	public void setValidation(
			Validation.ValidationType validationType,
			boolean ignoreBlank,
			Validation.OperatorType operatorType,
			boolean inCellDropDown,
			String formula1,
			String formula2,
			
			boolean showInput,
			String inputTitle,
			String inputMessage,
			
			boolean showError,
			Validation.AlertStyle alertStyle,
			String errorTitle,
			String errorMessage);
	
	
	/**
	 * Gets read-only {@link Validation} associated with the left-top cell
	 * of this range; return null if no associated validation.
	 * 
	 * @see #setValidation(Validation.ValidationType, boolean, Validation.OperatorType, boolean, String, String, boolean, String, String, boolean, Validation.AlertStyle, String, String)
	 * @since 3.5.0
	 */
	public List<Validation> getValidations();
	
	/**
	 * Delete the {@link Validation} associated in the specified range.
	 * @since 3.5.0 
	 */
	public void deleteValidation();

	//ZSS-742 Support API for input rich text
	/**
	 * Set into the left top cell of this Range the specified text in html format.
	 * @param html
	 * @since 3.6.0
	 */
	public void setCellRichText(String html);
	
	//ZSS-742
	/**
	 * Returns text in html format; null if not a rich text.
	 * @since 3.6.0
	 */
	public String getCellRichText();
	
	/**
	 * Get font if exists or create font if not exists from the book of this 
	 * range and return it.
	 * 
	 * @return font of the specified font metrics
	 * @since 3.6.0
	 * @deprecated As if release 5.3.0, replaced by {@link CellStyleHelper#builder(CellStyle)} to create an immutable
	 * cell style and font
	 */
	@Deprecated public Font getOrCreateFont(Font.Boldweight boldweight, Color color,
			int fontHeight, String fontName, boolean italic, boolean strikeout,
			Font.TypeOffset typeOffset, Font.Underline underline);

	/**
	 * Enforce evaluation(if not cached) and refresh UI of this range and its 
	 * dependent cells if the includeDependents is true. This method is equal
	 * to {@link Range#refresh(boolean includeDependents, boolean false, boolean true)}. 
	 * @since 3.6.0
	 */
	public void refresh(boolean includeDependants);
	
	/**
	 * Turn on(true)/off(false) of auto refresh of this range and return 
	 * previous on/off status; default is true. If you turn off auto refresh, you need to call {@link #notifyChange()} manually at the proper moment.
	 * 
	 * @param auto whether refresh automatically
	 * @return previous status
	 * @since 3.6.0
	 */
	public boolean setAutoRefresh(boolean auto);

	/**
	 * Update data model and refresh UI of this range and its dependent cells 
	 * if the argument includeDependents is true. 
	 * 
	 * Note that when you set parameter clearCache to true, the cached formula 
	 * result in data model will be cleared first.
	 * 
	 * If you set parameter enforceEval to true, data model associated with 
	 * this range will be evaluated immediately; otherwise will be 
	 * evaluated on demand.
	 *  
	 * @since 3.7.0
	 */
	public void refresh(boolean includeDependents, boolean clearCache, boolean enforceEval);
	
	/**
	 * Setup sheet's visibility; can be VISIBLE, HIDDEN, or VERY_HIDDEN.
	 * <ul>
	 * 	<li>VISIBLE: the sheet is visible</li>
	 *  <li>HIDDEN: the sheet is hidden but can be unhidden using UI</li>
	 *  <li>VERY_HIDDEN: the sheet is hidden but can be unhidden only via this API.</li>
	 * </ul>
	 * @param visible
	 * @see SheetVisible
	 * @since 3.7.0
	 */
	public void setSheetVisible(SheetVisible visible);

	/**
	 * Return the comment rich edit text of the left top cell of this Range.
	 * @return the rich edit text of this Range.
	 * @since 3.7.0
	 */
	public String getCommentRichText();

	/**
	 * Set comment rich text into the left top cell of this range; null to 
	 * delete it. 
	 * @since 3.7.0
	 */
	public void setCommentRichText(String html);

	/**
	 * Set comment visibility into the left top cell of this range; if no
	 * comment at the cell, it simply ignored.
	 * @param visible
	 * @since 3.7.0
	 */
	public void setCommentVisible(boolean visible);
	
	/**
	 * Returns whether the comment is always visible at the left top cell of
	 * this range; if no comment at the cell, it returns false.
	 * @return
	 * @since 3.7.0
	 */
	public boolean isCommentVisible();

	/**
	 * Notify this range has been changed on the specifield attribute.
	 * @since 3.8.0
	 */
	public void notifyChange(CellAttribute cellAttr);

	/**
     * Set(Rename) the name of the Name(or Table) as specified in this 
     * Range(a Book or a Sheet).
	 * @param namename name of the Name(or Table)
	 * @param newname new name for the Name(or Table) 
	 * @since 3.8.0
	 */
	public void setNameName(String namename, String newname);
	
	/**
	 * Save the specified value into this range as a String no matter the value 
	 * is prefixed with '=' or not. 
	 * @param value
	 * @since 3.8.0
	 */
	public void setStringValue(String value);

	/**
	 * Returns the merged region of the left top cell of this Range if any; 
	 * otherwise, returns null if not a merged region. 
	 * @return the merged region of the left top cell of this Range if any
	 * @since 3.8.3
	 */
	public CellRegion getMergedRegion();

	/**
	 * Returns the region which contains data(including chart, picture, and 
	 * fills) in the sheet of this Range; return null if empty sheet.
	 * @return the region which contains data(including chart, picture, and 
	 * fills) in the sheet of this Range; return null if empty sheet.
	 * @since 3.8.3
	 */
	public CellRegion getDataRegion();

	/**
	 * Clone from the specified source sheet to the owner book of this range.
	 * The book can be different between the owner book of this range and
	 * of the specified source sheet.
	 *
	 * @param name the name of the new created sheet; null would try to use
	 * the sheet name of the specified source sheet; if the same sheet name 
	 * already used in the owner book of this range then use default "SheetX" 
	 * name where X is the next sheet number of the owner book of this range.
	 * @param sheet the source sheet to clone from to the owner book of this 
	 * range.
	 * @return the new created sheet
	 * @since 3.9.0
	 * @see #cloneSheetFrom(String, Sheet, boolean)
	 */
	public Sheet cloneSheetFrom(String name, Sheet sheet);

	/**
	 * Clone from the specified source sheet to the owner book of this range.
	 * The book can be different between the owner book of this range and
	 * of the specified source sheet.
	 *
	 * @param name the name of the new created sheet; null would try to use
	 * the sheet name of the specified source sheet; if the same sheet name
	 * already used in the owner book of this range then use default "SheetX"
	 * name where X is the next sheet number of the owner book of this range.
	 * @param sheet the source sheet to clone from to the owner book of this
	 * range.
	 * @param valueOnly true to copy the cell value only
	 * @return the new created sheet
	 * @since 6.0.0
	 */
	public Sheet cloneSheetFrom(String name, Sheet sheet, boolean valueOnly);

	/**
	 * Returns a z-order cell range stream which treats a merged cell as one range.
	 * For example,
	 * <pre><code>
	 *    //  Data:
	 *    //     | A | B | C | D | E |
	 *    //  ---+---+---+---+---+---+
	 *    //   1 | 1 |       | 3 |   |
	 *    //  ---+---+       +---+---+
	 *    //   2 |   |       |   |   |
	 *    //  ---+   +       + 5 +---+
	 *    //   3 | 4 |   2   |   |   |
	 *    //  ---+   +       +---+---+
	 *    //   4 |   |       | 6 |   |
	 *    //  ---+---+       +---+---+
	 *    //   5 | 7 |       | 8 |   |
	 *    //  ---+---+-------+---+---+
	 *
	 *    Stream<Range> zOrder = </Range>Ranges.range("A1:D5").zOrderStream();
	 * 	  zOrder.map(Range::getCellValue)
	 * 	        .map(Object::toString)
	 * 	        .collect(Collectors.joining(", "));
	 * 	  // Result: 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0
	 * </code></pre>
	 * @since 5.6.0
	 */
	public Stream<Range> zOrderStream();

	/**
	 * Returns a n-order cell range stream which treats a merged cell as one range.
	 * For example,
	 * <pre><code>
	 *     //  Data:
	 *     //     | A | B | C | D | E |
	 *     //  ---+---+---+---+---+---+
	 *     //   1 | 1 |       | 3 |   |
	 *     //  ---+---+       +---+---+
	 *     //   2 |   |       |   |   |
	 *     //  ---+   +       + 5 +---+
	 *     //   3 | 4 |   2   |   |   |
	 *     //  ---+   +       +---+---+
	 *     //   4 |   |       | 6 |   |
	 *     //  ---+---+       +---+---+
	 *     //   5 | 7 |       | 8 |   |
	 *     //  ---+---+-------+---+---+
	 *
	 *    Stream<Range> nOrder = </Range>Ranges.range("A1:D5").nOrderStream();
	 * 	  nOrder.map(Range::getCellValue)
	 * 	        .map(Object::toString)
	 * 	        .collect(Collectors.joining(", "));
	 *
	 * 	  // Result: 1.0, 4.0, 7.0, 2.0, 3.0, 5.0, 6.0, 8.0
	 * </code></pre>
	 * @since 5.6.0
	 */
	public Stream<Range> nOrderStream();

	// KEIKAI-396
	/**
	 * Returns Range(s) that are the direct precedents of this cell Range.
	 * If this Range covers more than one cell, the left-top one is used.
	 * @since 5.6.0
	 */
	public Set<Range> getDirectPrecedents();

	// KEIKAI-396
	/**
	 * Returns all Range(s) that are the direct or indirect precedents of this cell Range.
	 * If this Range covers more than one cell, the left-top one is used.
	 * @since 5.6.0
	 */
	public Set<Range> getPrecedents();

	// KEIKAI-396
	/**
	 * Returns Range(s) that are the direct dependents of this cell Range.
	 * If this Range covers more than one cell, the left-top one is used.
	 * @since 5.6.0
	 */
	public Set<Range> getDirectDependents();

	// KEIKAI-396
	/**
	 * Returns all Range(s) that are the direct or indirect dependents of this cell Range.
	 * If this Range covers more than one cell, the left-top one is used.
	 * @since 5.6.0
	 */
	public Set<Range> getDependents();

	/**
	 * Trace the precedents of this range.
	 * @since 5.6.0
	 */
	public void tracePrecedents();

	/**
	 * Trace the dependents of this range.
	 * @since 5.6.0
	 */
	public void traceDependents();

	/**
	 * Clears the current tracing.
	 * @since 5.6.0
	 */
	public void clearTrace();

	/**
	 * Returns all group ranges for rows, which intersect
	 * with this range.
	 * @since 6.0.0
	 * @return an empty list if not matched.
	 */
	public List<Range> getRowGroups();

	/**
	 * Returns all group ranges matched the given outline level for rows, which intersect
	 * with this range.
	 * @param outlineLevel a valid value in a range is from 1 to 7 (inclusive).
	 * @since 6.0.0
	 * @return an empty list if not matched.
	 */
	public List<Range> getRowGroups(int outlineLevel);

	/**
	 * Returns all group ranges matched between fromOutlineLevel and toOutlineLevel
	 * for rows, which intersect with this range.
	 * @param fromOutlineLevel 1~7 (inclusive)
	 * @param toOutlineLevel 2~8 (exclusive)
	 * @since 6.0.0
	 * @return an empty list if not matched.
	 */
	public List<Range> getRowGroups(int fromOutlineLevel, int toOutlineLevel);

	/**
	 * Returns all group ranges for columns, which intersect
	 * with this range.
	 * @since 6.0.0
	 * @return an empty list if not matched.
	 */
	public List<Range> getColumnGroups();

	/**
	 * Returns all group ranges matched the given outline level for columns, which intersect
	 * with this range.
	 * @param outlineLevel a valid value in a range is from 1 to 7 (inclusive).
	 * @since 6.0.0
	 * @return an empty list if not matched.
	 */
	public List<Range> getColumnGroups(int outlineLevel);

	/**
	 * Returns all group ranges matched between fromOutlineLevel and toOutlineLevel
	 * for columns, which intersect with this range.
	 * @param fromOutlineLevel 1~7 (inclusive)
	 * @param toOutlineLevel 2~8 (exclusive)
	 * @since 6.0.0
	 * @return an empty list if not matched.
	 */
	public List<Range> getColumnGroups(int fromOutlineLevel, int toOutlineLevel);

	/**
	 * Sets to be true to display summary group as below or to be false to display
	 * summary group as above for rows.
	 * <p><b>Note:</b> One setting to affect whole sheet, if the value is up-side-down
	 * against the previous value, it will clear all outline level for rows and columns.</p>
	 * <p>Default: {@code true}</p>
	 * @param summaryBelow true to display as below for rows.
	 * @since 6.0.0
	 */
	public void setSummaryBelow(boolean summaryBelow);

	/**
	 * Returns whether to display group summary as below for rows.
	 * <p>Default: {@code true}</p>
	 * @since 6.0.0
	 */
	public boolean isSummaryBelow();

	/**
	 * Sets to be true to display summary group as right or to be false to display
	 * summary group as left for columns.
	 * <p><b>Note:</b> One setting to affect whole sheet, if the value is up-side-down
	 * against the previous value, it will clear all outline level for rows and columns.</p>
	 * <p>Default: {@code true}</p>
	 * @param summaryRight true to display as right for columns.
	 * @since 6.0.0
	 */
	public void setSummaryRight(boolean summaryRight);

	/**
	 * Returns whether to display group summary as right for columns.
	 * <p>Default: {@code true}</p>
	 * @since 6.0.0
	 */
	public boolean isSummaryRight();

	/**
	 * Collapses one group by matching with the range at a time (not affected its descendant).
	 * @since 6.0.0
	 */
	public void collapse();

	/**
	 * Expands one group by matching with the range at a time (not affected its descendant).
	 * @since 6.0.0
	 */
	public void expand();

	/**
	 * Clears all outline level for all groups that intersect with the range.
	 * {@code 2:4} for rows and {@code C:E} for columns
	 */
	public void clearOutlineLevel();

	/**
	 * Returns whether the group by matching with the range has collapsed or not.
	 * Otherwise, returns false instead.
	 * @since 6.0.0
	 */
	public boolean isCollapsed();

	/**
	 * Returns the outline of the group by matching with the range.
	 * Otherwise, returns 0 instead.
	 * @since 6.0.0
	 */
	public int getOutlineLevel();

	/**
	 * Groups the rows or columns by matching the range, which increases their outline
	 * level.
	 * <p><b>Note:</b> The maximum level is up to 7.
	 * @since 6.0.0
	 */
	public void group();

	/**
	 * Ungroups the rows or columns by matching the range, which decreases their outline
	 * level.
	 * <p><b>Note:</b> The minimum level is down to 0.
	 * @since 6.0.0
	 */
	public void ungroup();
}
