/* CellStyle.java

{{IS_NOTE
	Purpose:
		
	Description:
		
	History:
		2013/5/1 , Created by dennis
}}IS_NOTE

Copyright (C) 2013 Potix Corporation. All Rights Reserved.

{{IS_RIGHT
}}IS_RIGHT
*/
package io.keikai.api.model;


/**
 * This interface provides access to "style" part of a cell including alignment, border, font, formant, and color.
 * @author dennis
 * @since 3.0.0
 */
public interface CellStyle {

	/**
	 * The fill pattern
	 * @since 3.7.0
	 */
	public enum FillPattern {
		NONE, //NO_FILL
		SOLID, //SOLID_FOREGROUND
		MEDIUM_GRAY, //FINE_DOTS
		DARK_GRAY, //ALT_BARS
		LIGHT_GRAY, //SPARSE_DOTS
		DARK_HORIZONTAL, //THICK_HORZ_BANDS
		DARK_VERTICAL, //THICK_VERT_BANDS
		DARK_DOWN, //THICK_BACKWARD_DIAG
		DARK_UP, //THICK_FORWARD_DIAG
		DARK_GRID, //BIG_SPOTS
		DARK_TRELLIS, //BRICKS
		LIGHT_HORIZONTAL, //THIN_HORZ_BANDS
		LIGHT_VERTICAL, //THIN_VERT_BANDS
		LIGHT_DOWN, //THIN_BACKWARD_DIAG
		LIGHT_UP, //THIN_FORWARD_DIAG
		LIGHT_GRID, //SQUARES
		LIGHT_TRELLIS, //DIAMONDS
		GRAY125, //LESS_DOTS 
		GRAY0625 //LEAST_DOTS
	}

	/**
	 * The horizontal alignment
	 *
	 */
	public enum Alignment {
		GENERAL, LEFT, CENTER, RIGHT, FILL, JUSTIFY, CENTER_SELECTION
	}

	/**
	 * The vertical alignment
	 */
	public enum VerticalAlignment {
		TOP, CENTER, BOTTOM, JUSTIFY
	}

	/**
	 * The border type
	 */
	public enum BorderType {
		NONE, THIN, MEDIUM, DASHED, HAIR, THICK, DOUBLE, DOTTED, MEDIUM_DASHED, DASH_DOT, MEDIUM_DASH_DOT, DASH_DOT_DOT, MEDIUM_DASH_DOT_DOT, SLANTED_DASH_DOT;
	}

	/**
	 * @return the font
	 */
	public Font getFont();
	
	/**
	 * 
	 * @return fill foreground color
	 * @since 3.5.0
	 */
	public Color getFillColor();
	
	/**
	 * 
	 * @return fill background color
	 * @since 3.6.0
	 */
	public Color getBackColor();

	/**
	 * Gets the fill/background pattern <br/>
	 * @return the fill pattern
	 */
	public FillPattern getFillPattern();

	/**
	 * Gets the horizontal alignment <br/>
	 * @return the horizontal alignment
	 */
	public Alignment getAlignment();

	/**
	 * Gets vertical alignment <br/>
	 * @return
	 */
	public VerticalAlignment getVerticalAlignment();

	/**
	 * @return true if wrap-text
	 */
	public boolean isWrapText();

	/**
	 * @return left border
	 */
	public BorderType getBorderLeft();

	/**
	 * @return top border
	 */
	public BorderType getBorderTop();

	/**
	 * @return right border
	 */
	public BorderType getBorderRight();

	/**
	 * @return bottom border
	 */
	public BorderType getBorderBottom();

	/**
	 * @return top border color
	 */
	public Color getBorderTopColor();

	/** 
	 * @return left border color
	 */
	public Color getBorderLeftColor();

	/**
	 * 
	 * @return bottom border color
	 */
	public Color getBorderBottomColor();

	/**
	 * @return right border color
	 */
	public Color getBorderRightColor();
	
	/**
	 * @return data format
	 */
	public String getDataFormat();
	
	/**
	 * 
	 * @return true if locked
	 */
	public boolean isLocked();
	
	/**
	 * 
	 * @return true if hidden
	 */
	public boolean isHidden();

	/**
	 * Cell style builder to create the cell style
	 * @since 5.3.0
	 */
	interface Builder {
		/**
		 * Sets the font name.
		 * @param name font name, e.g. "Calibri"
		 */
		Builder fontName(String name);

		/**
		 * Sets the font color.
		 * @param color font color
		 */
		Builder fontColor(Color color);

		/**
		 * Sets the font bold style.
		 * @param boldweight font bold weight
		 */
		Builder fontBoldweight(Font.Boldweight boldweight);

		/**
		 * Sets if the font is italic.
		 * @param italic font italic
		 */
		Builder fontItalic(boolean italic);

		/**
		 * Sets if the font is strikeout.
		 * @param strikeout font strikeout
		 */
		Builder fontStrikeout(boolean strikeout);

		/**
		 * Sets the font underline style.
		 * @param underline font underline
		 */
		Builder fontUnderline(Font.Underline underline);

		/**
		 * Sets the font height
		 * @param height font height
		 */
		Builder fontHeight(int height);

		/**
		 * Sets the font height in point
		 * @param point font height in point
		 */
		Builder fontHeightInPoint(int point);

		/**
		 * Sets background color of a fill.
		 * @param color background color
		 */
		Builder backColor(Color color);

		/**
		 * Sets foreground color of a fill.
		 * @param color fill color
		 */
		Builder fillColor(Color color);

		/**
		 * Sets the fill/background pattern.
		 * Note: Spreadsheet (UI display) supports only {@link FillPattern#NONE} and {@link FillPattern#SOLID}
		 * (Other pattern will be display as {@link FillPattern#SOLID),
		 * However you can still set another pattern, the data will still be kept when export.
		 * @param pattern fill pattern
		 */
		Builder fillPattern(FillPattern pattern);

		/**
		 * Sets horizontal alignment.
		 * Note: Spreadsheet(UI display) supports only {@link Alignment#LEFT}, {@link Alignment#CENTER}, {@link Alignment#RIGHT}
		 * ({@link Alignment#CENTER_SELECTION} will be display as {@link Alignment#CENTER}, Other alignment will be display as  {@link Alignment#LEFT}),
		 * However you can still set another alignment, the data will still be kept when export.
		 * @param alignment cell horizontal alignment
		 */
		Builder alignment(Alignment alignment);

		/**
		 * Sets vertical alignment.
		 * Note: Spreadsheet(UI display) supports only {@link VerticalAlignment#TOP}, {@link VerticalAlignment#CENTER}, {@link VerticalAlignment#TOP},
		 * (Other alignment will be display as  {@link VerticalAlignment#BOTTOM}),
		 * @param alignment cell vertical alignment
		 */
		Builder verticalAlignment(VerticalAlignment alignment);

		/**
		 * Sets wrap text.
		 * @param wraptext wrap text
		 */
		Builder wrapText(boolean wraptext);

		/**
		 * Sets left border.
		 * Note: Spreadsheet(UI display) only supports {@link BorderType#NONE}, {@link BorderType#THIN}, {@link BorderType#DOTTED} and {@link BorderType#HAIR},
		 ({@link BorderType#DASH_DOT} will be display as {@link BorderType#DOTTED}, Other type will be display as {@link BorderType#THIN}),
		 * However you can still set another border type, the data will still be kept when export.
		 * @param borderType left border type
		 */
		Builder borderLeft(BorderType borderType);

		/**
		 * Sets top border.
		 * Note: Spreadsheet(UI display) only supports {@link BorderType#NONE}, {@link BorderType#THIN}, {@link BorderType#DOTTED} and {@link BorderType#HAIR},
		 ({@link BorderType#DASH_DOT} will be display as {@link BorderType#DOTTED}, Other type will be display as {@link BorderType#THIN}),
		 * However you can still set another border type, the data will still be kept when export.
		 * @param borderType top border type
		 */
		Builder borderTop(BorderType borderType);

		/**
		 * Sets right border.
		 * Note: Spreadsheet(UI display) only supports {@link BorderType#NONE}, {@link BorderType#THIN}, {@link BorderType#DOTTED} and {@link BorderType#HAIR},
		 ({@link BorderType#DASH_DOT} will be display as {@link BorderType#DOTTED}, Other type will be display as {@link BorderType#THIN}),
		 * However you can still set another border type, the data will still be kept when export.
		 * @param borderType right border type
		 */
		Builder borderRight(BorderType borderType);

		/**
		 * Sets bottom border.
		 * Note: Spreadsheet(UI display) only supports {@link BorderType#NONE}, {@link BorderType#THIN}, {@link BorderType#DOTTED} and {@link BorderType#HAIR},
		 * ({@link BorderType#DASH_DOT} will be display as {@link BorderType#DOTTED}, Other type will be display as {@link BorderType#THIN}),
		 * However you can still set another border type, the data will still be kept when export.
		 * @param borderType bottom border type
		 */
		Builder borderBottom(BorderType borderType);

		/**
		 * Sets top border color.
		 * you could use {@link io.keikai.api.Range.CellStyleHelper#createColorFromHtmlColor(String)} to create a {@link Color}
		 * @param color top border color
		 */
		Builder borderTopColor(Color color);

		/**
		 * Sets left border color.
		 * you could use {@link io.keikai.api.Range.CellStyleHelper#createColorFromHtmlColor(String)} to create a {@link Color}
		 * @param color left border color
		 */
		Builder borderLeftColor(Color color);

		/**
		 * Sets bottom border color.
		 * you could use {@link io.keikai.api.Range.CellStyleHelper#createColorFromHtmlColor(String)} to create a {@link Color}
		 * @param color bottom border color
		 */
		Builder borderBottomColor(Color color);

		/**
		 * Sets right border color.
		 * you could use {@link io.keikai.api.Range.CellStyleHelper#createColorFromHtmlColor(String)} to create a {@link Color}
		 * @param color right border color
		 */
		Builder borderRightColor(Color color);

		/**
		 * Sets the data format
		 * @param format data format
		 */
		Builder dataFormat(String format);

		/**
		 * Sets if the cell should be locked.
		 * @param locked cell locked
		 */
		Builder locked(boolean locked);

		/**
		 * Sets if the cell formula should be hidden.
		 * @param hidden cell formula hidden
		 */
		Builder hidden(boolean hidden);

		/**
		 * Returns an immutable cell style object.
		 */
		CellStyle build();
	}
}
