package io.keikai.api.model.impl;

import io.keikai.api.UnitUtil;
import io.keikai.api.model.CellStyle;
import io.keikai.api.model.Color;
import io.keikai.api.model.Font;
import io.keikai.model.SBook;
import io.keikai.model.SCellStyle;
import io.keikai.model.SCellStyleBuilder;
import io.keikai.model.SFontBuilder;

import java.io.Serializable;

public class CellStyleBuilderImpl implements CellStyle.Builder, Serializable {

    private static final long serialVersionUID = 8161534520746689556L;

    private final SBook book;
    private final SCellStyleBuilder builder;
    private SFontBuilder fontBuilder;

    public CellStyleBuilderImpl(SBook book, SCellStyle src) {
        this.book = book;
        builder = book.addCellStyle(src);
    }

    @Override
    public CellStyle.Builder fontName(String name) {
        getOrCreateFontBuilder().name(name);
        return this;
    }

    @Override
    public CellStyle.Builder fontColor(Color color) {
        getOrCreateFontBuilder().color(color.getHtmlColor());
        return this;
    }

    @Override
    public CellStyle.Builder fontBoldweight(Font.Boldweight boldweight) {
        getOrCreateFontBuilder().boldweight(EnumUtil.toFontBoldweight(boldweight));
        return this;
    }

    @Override
    public CellStyle.Builder fontItalic(boolean italic) {
        getOrCreateFontBuilder().italic(italic);
        return this;
    }

    @Override
    public CellStyle.Builder fontStrikeout(boolean strikeout) {
        getOrCreateFontBuilder().strikeout(strikeout);
        return this;
    }

    @Override
    public CellStyle.Builder fontUnderline(Font.Underline underline) {
        getOrCreateFontBuilder().underline(EnumUtil.toFontUnderline(underline));
        return this;
    }

    @Override
    public CellStyle.Builder fontHeight(int height) {
        return fontHeightInPoint(UnitUtil.twipToPoint(height));
    }

    @Override
    public CellStyle.Builder fontHeightInPoint(int point) {
        getOrCreateFontBuilder().heightPoints(point);
        return this;
    }

    @Override
    public CellStyle.Builder backColor(Color color) {
        builder.backColor(color.getHtmlColor());
        return this;
    }

    @Override
    public CellStyle.Builder fillColor(Color color) {
        builder.fillColor(color.getHtmlColor());
        return this;
    }

    @Override
    public CellStyle.Builder fillPattern(CellStyle.FillPattern pattern) {
        builder.fillPattern(EnumUtil.toStyleFillPattern(pattern));
        return this;
    }

    @Override
    public CellStyle.Builder alignment(CellStyle.Alignment alignment) {
        builder.alignment(EnumUtil.toStyleAlignemnt(alignment));
        return this;
    }

    @Override
    public CellStyle.Builder verticalAlignment(CellStyle.VerticalAlignment alignment) {
        builder.verticalAlignment(EnumUtil.toStyleVerticalAlignemnt(alignment));
        return this;
    }

    @Override
    public CellStyle.Builder wrapText(boolean wrapText) {
        builder.wrapText(wrapText);
        return this;
    }

    @Override
    public CellStyle.Builder borderLeft(CellStyle.BorderType borderType) {
        builder.borderLeft(EnumUtil.toStyleBorderType(borderType));
        return this;
    }

    @Override
    public CellStyle.Builder borderTop(CellStyle.BorderType borderType) {
        builder.borderTop(EnumUtil.toStyleBorderType(borderType));
        return this;
    }

    @Override
    public CellStyle.Builder borderRight(CellStyle.BorderType borderType) {
        builder.borderRight(EnumUtil.toStyleBorderType(borderType));
        return this;
    }

    @Override
    public CellStyle.Builder borderBottom(CellStyle.BorderType borderType) {
        builder.borderBottom(EnumUtil.toStyleBorderType(borderType));
        return this;
    }

    @Override
    public CellStyle.Builder borderTopColor(Color color) {
        builder.borderTopColor(color.getHtmlColor());
        return this;
    }

    @Override
    public CellStyle.Builder borderLeftColor(Color color) {
        builder.borderLeftColor(color.getHtmlColor());
        return this;
    }

    @Override
    public CellStyle.Builder borderBottomColor(Color color) {
        builder.borderBottomColor(color.getHtmlColor());
        return this;
    }

    @Override
    public CellStyle.Builder borderRightColor(Color color) {
        builder.borderRightColor(color.getHtmlColor());
        return this;
    }

    @Override
    public CellStyle.Builder dataFormat(String format) {
        builder.dataFormat(format);
        return this;
    }

    @Override
    public CellStyle.Builder locked(boolean locked) {
        builder.locked(locked);
        return this;
    }

    @Override
    public CellStyle.Builder hidden(boolean hidden) {
        builder.hidden(hidden);
        return this;
    }

    @Override
    public CellStyle build() {
        if (fontBuilder != null) {
            builder.font(fontBuilder.build());
        }
        return new CellStyleImpl(new SimpleRef<>(book), new SimpleRef<>(builder.build()));
    }

    private SFontBuilder getOrCreateFontBuilder() {
        if (fontBuilder == null) {
            fontBuilder = book.createFont();
        }
        return fontBuilder;
    }
}
