/* ToolbarButton.java

	Purpose:
		
	Description:
		
	History:
		11:57 AM 2019/11/12, Created by jumperchen

Copyright (C) 2019 Potix Corporation. All Rights Reserved.
*/
package io.keikai.ui;

import org.zkoss.json.JSONAware;
import org.zkoss.json.JSONObject;
import org.zkoss.zk.ui.Execution;
import org.zkoss.zk.ui.Executions;

import java.io.Serializable;

/**
 * A toolbar button for Spreadsheet component (Immutable).
 * <p> If the button's key is duplicated in {@link AuxAction#getAction()},
 * then the {@link #getTemplateURI()} will be ignored and using the default template instead.
 * Due to this limitation, you can remove the default button's key and then add
 * a new toolbar button with an extra key which provides a different behavior implementation
 * for it.</p>
 * @author jumperchen
 * @since 5.0.0
 */
public class ToolbarButton implements JSONAware, Serializable {
	final private String key;
	/*package*/ String iconURI;
	/*package*/ String iconClass;
	/*package*/ String templateURI;
	/*package*/ String tooltip;
	/*package*/ String label;

	// use Builder to create this instance.
	private ToolbarButton(String key) {
		this.key = key;
	}

	/**
	 * Returns the button key.
	 * @return
	 */
	public String getKey() {
		return key;
	}

	/**
	 * Returns the button label.
	 * @return
	 */
	public String getLabel() {
		return label;
	}
	/**
	 * Returns the icon uri, if any.
	 * @return
	 */
	public String getIconURI() {
		return iconURI;
	}
	/**
	 * Returns the icon class, if any.
	 * @return
	 * @since 5.2.0
	 */
	public String getIconClass() {
		return iconClass;
	}

	/**
	 * Returns the tooltip text, if any.
	 * @return
	 */
	public String getTooltip() {
		return tooltip;
	}

	/**
	 * Returns the template uri, if any.
	 * @return
	 */
	public String getTemplateURI() {
		return templateURI;
	}

	public String toJSONString() {
		Execution current = Executions.getCurrent();
		JSONObject map = new JSONObject();
		map.put("key", key);
		if (iconURI != null) {
			map.put("iconURI", current != null ? current.encodeURL(iconURI) : iconURI);
		}
		if (iconClass != null) {
			map.put("iconClass", current != null ? current.encodeURL(iconClass) : iconClass);
		}
		if (tooltip != null) {
			map.put("tooltip", tooltip);
		}
		if (label != null) {
			map.put("label", label);
		}
		if (templateURI != null) {
			map.put("templateURI", current != null ? current.encodeURL(templateURI) : templateURI);
		}
		return map.toJSONString();
	}

	/**
	 * Toolbar button builder
	 */
	public static class Builder {
		ToolbarButton _btn;
		private Builder(String key) {
			_btn = new ToolbarButton(key);
		}

		/**
		 * Creates the builder with the given key.
		 * @param key the button key.
		 * @return
		 */
		public static Builder create(String key) {
			if (key == null) throw new NullPointerException("The button key cannot be null!");
			return new Builder(key);
		}

		/**
		 * Sets the icon URI to the button.
		 * @param iconURI
		 * @return
		 */
		public Builder withIconURI(String iconURI) {
			_btn.iconURI = iconURI;
			return this;
		}
		/**
		 * Sets the icon class to the button.
		 * Note that, if no template specified, yet iconClass and iconURI is specified at the same time, iconURI would be ignored.
		 * @param iconClass
		 * @return
		 * @since 5.2.0
		 */
		public Builder withIconClass(String iconClass) {
			_btn.iconClass = iconClass;
			return this;
		}

		/**
		 * Sets the tooltip to the button.
		 * @param tooltip
		 * @return
		 */
		public Builder withTooltip(String tooltip) {
			_btn.tooltip = tooltip;
			return this;
		}

		/**
		 * Sets the label to the button.
		 * @param label
		 * @return
		 */
		public Builder withLabel(String label) {
			_btn.label = label;
			return this;
		}

		/**
		 * Sets the template URI to the button.
		 * @param templateURI
		 * @return
		 */
		public Builder withTemplateURI(String templateURI) {
			_btn.templateURI = templateURI;
			return this;
		}

		/**
		 * Builds the toolbar button instance.
		 * @return
		 */
		public ToolbarButton build() {
			return _btn;
		}
	}
}
