/* MergeMatrixHelper.java

{{IS_NOTE
	Purpose:
		
	Description:
		
	History:
		Mar 18, 2008 1:08:11 PM     2008, Created by Dennis.Chen
}}IS_NOTE

Copyright (C) 2007 Potix Corporation. All Rights Reserved.

{{IS_RIGHT
	This program is distributed under GPL Version 2.0 in the hope that
	it will be useful, but WITHOUT ANY WARRANTY.
}}IS_RIGHT
*/
package io.keikai.ui.impl;

import java.io.Serializable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.StreamSupport;

import com.github.davidmoten.rtree2.Entry;
import com.github.davidmoten.rtree2.RTree;
import com.github.davidmoten.rtree2.geometry.Geometries;
import com.github.davidmoten.rtree2.geometry.Rectangle;
import io.keikai.model.CellRegion;
import io.keikai.util.Ref;

/**
 * Each sheet has its own MergeMatrixHelper and it manages all merged cells.
 * @author Dennis.Chen
 * @author jumperchen
 */
public class MergeMatrixHelper implements Serializable {
	private static final long serialVersionUID = -8360458859084652393L;
	
	private List<MergedRect> _mergeRanges = new LinkedList<MergedRect>();
	
	private int _frozenRow;
	private int _frozenCol;//don't care this for now.
	private SequenceId _mergeId = new SequenceId(0,1);
	private transient RTree<MergedRect, Rectangle> _rtree = RTree
			.create();
	/**
	 * @param mergeRange List of merge range
	 * @param frozenRow
	 * @param frozenCol
	 */
	public MergeMatrixHelper(List<CellRegion> mergeRange,int frozenRow,int frozenCol){
		mergeRange.forEach(region -> _mergeRanges
				.add(new MergedRect(_mergeId.next(), region.column, region.row,
						region.lastColumn, region.lastRow)));
		this._frozenRow = frozenRow;
		this._frozenCol = frozenCol;

		rebuildRTree();
	}

	private synchronized void writeObject(java.io.ObjectOutputStream s)
			throws java.io.IOException {
		s.defaultWriteObject();
	}

	private void readObject(java.io.ObjectInputStream s)
			throws java.io.IOException, ClassNotFoundException {
		s.defaultReadObject();

		rebuildRTree();
	}

	private void rebuildRTree() {
		_rtree = RTree.create();
		if (!this._mergeRanges.isEmpty()) {
			for (MergedRect rect : _mergeRanges) {
				_rtree = _rtree.add(rect, area2Rect(rect));
			}
		}
	}

	private static Rectangle area2Rect(MergedRect rect) {
		return Geometries.rectangle(rect.getColumn(), rect.getRow(), rect.getLastColumn(), rect.getLastRow());
	}
	
	
	public void update(int frozenRow,int frozenCol){
		this._frozenRow = frozenRow;
		this._frozenCol = frozenCol;
	}
	
	/**
	 * Check is (row,col) in one of merge range's left-top 
	 */
	public boolean isMergeRangeLeftTop(int row, int col) {
		Iterator<Entry<MergedRect, Rectangle>> lists = _rtree
				.search(Geometries.point(col, row)).iterator();
		while (lists.hasNext()) {
			MergedRect mergedRect = lists.next().value();
			if (mergedRect.getColumn() == col && mergedRect.getRow() == row) {
				return true;
			}
		}
		return false;
	}
	
	/**
	 * Get a merged range which contains cell(row,col)
	 */
	public MergedRect getMergeRange(int row,int col){
		Iterator<Entry<MergedRect, Rectangle>> lists = _rtree
				.search(Geometries.point(col, row)).iterator();
		if (lists.hasNext()) {
			return lists.next().value();
		}
		return null;
	}
	
	/**
	 * Get all merged range. 
	 * @return a list which contains all merged range
	 */
	public List<MergedRect> getRanges(){
		return _mergeRanges;
	}
	
	/**
	 * Returns {@link MergedRect} MergedRect by id
	 * 
	 * @param id
	 * @return MergedRect
	 */
	public MergedRect getMergedRect(int id) {
		for (MergedRect r : _mergeRanges) {
			if (r.getId() == id) {
				return r;
			}
		}
		return null;
	}
	
	/**
	 * Get merged range which contains col
	 * @param col column index
	 * @return a list which contains merged range
	 */
	public Set getRangesByColumn(int col) {
		Iterator<Entry<MergedRect, Rectangle>> lists = _rtree
				.search(Geometries.rectangle(col, 0, col, Ref.MAX_ROW - 1)).iterator();
		return StreamSupport.stream(((Iterable<Entry<MergedRect, Rectangle>>)() -> lists).spliterator(), false).map(entry -> entry.value()).collect(
				Collectors.toSet());
	}

	public int getRightConnectedColumn(int col, int top, int bottom) {
		Iterator<Entry<MergedRect, Rectangle>> lists = _rtree
				.search(Geometries.rectangle(col, Math.max(_frozenRow, top), col, bottom)).iterator();

		while (lists.hasNext()) {
			MergedRect rect = lists.next().value();
			if (rect.getLastColumn() > col && rect.getColumn() <= col) {
				col = rect.getLastColumn();
			}
		}
		return col;
	}

	public int getLeftConnectedColumn(int col, int top, int bottom) {
		Iterator<Entry<MergedRect, Rectangle>> lists = _rtree
				.search(Geometries.rectangle(col, Math.max(_frozenRow, top), col, bottom)).iterator();

		while (lists.hasNext()) {
			MergedRect rect = lists.next().value();
			if (rect.getColumn() < col && rect.getLastColumn() >= col) {
				col = rect.getColumn();
			}
		}
		return col;
	}

	public int getBottomConnectedRow(int row, int left, int right) {
		Iterator<Entry<MergedRect, Rectangle>> lists = _rtree
				.search(Geometries.rectangle(Math.max(_frozenCol, left), row, right, row)).iterator();

		while (lists.hasNext()) {
			MergedRect rect = lists.next().value();
			if (rect.getLastRow() > row && rect.getRow() <= row) {
				row = rect.getLastRow();
			}
		}
		return row;
	}

	public int getTopConnectedRow(int row, int left, int right) {
		Iterator<Entry<MergedRect, Rectangle>> lists = _rtree
				.search(Geometries.rectangle(Math.max(_frozenCol, left), row, right, row)).iterator();

		while (lists.hasNext()) {
			MergedRect rect = lists.next().value();
			if (rect.getRow() < row && rect.getLastRow() >= row) {
				row = rect.getRow();
			}
		}
		return row;
	}

	public void updateMergeRange(int oleft, int otop, int oright, int obottom,
			int left, int top, int right, int bottom, Set toadd, Set torem) {
		for (int i = otop; i <= obottom; i++) {
			MergedRect mblock = getMergeRange(i, oleft);
			if (mblock != null) {
				torem.add(mblock);
				_mergeRanges.remove(mblock);
				_rtree = _rtree.delete(mblock, area2Rect(mblock));
			}
		}

		final MergedRect mblock = new MergedRect(_mergeId.next(), left, top,
				right, bottom);
		toadd.add(mblock);
		_mergeRanges.add(mblock);
		_rtree = _rtree.add(mblock, area2Rect(mblock));
	}

	//ZSS-1168
	public void deleteMergeRange(int left, int top, int right, int bottom, Set torem) {
		final MergedRect mblock = getMergeRange(top, left);
		if (mblock != null) {
			torem.add(mblock);
			_mergeRanges.remove(mblock);
			_rtree = _rtree.delete(mblock, area2Rect(mblock));
		}
	}

	//ZSS-1168
	public void addMergeRange(int left, int top, int right, int bottom, Set toadd, Set torem) {
		MergedRect mblock = getMergeRange(top, left);
		if (mblock != null) {
			torem.add(mblock);
			_mergeRanges.remove(mblock);
			_rtree = _rtree.delete(mblock, area2Rect(mblock));
		}

		mblock = new MergedRect(_mergeId.next(), left, top, right, bottom);
		toadd.add(mblock);
		_mergeRanges.add(mblock);
		_rtree = _rtree.add(mblock, area2Rect(mblock));
	}
	
}
